#!/usr/bin/env python3

import logging
import struct
import time
import sys

import click

from pymobiledevice3.lockdown import LockdownClient

"""
struct pcap_hdr_s {
        guint32 magic_number;   /* magic number */
        guint16 version_major;  /* major version number */
        guint16 version_minor;  /* minor version number */
        gint32  thiszone;       /* GMT to local correction */
        guint32 sigfigs;        /* accuracy of timestamps */
        guint32 snaplen;        /* max length of captured packets, in octets */
        guint32 network;        /* data link type */
} pcap_hdr_t;
typedef struct pcaprec_hdr_s {
        guint32 ts_sec;         /* timestamp seconds */
        guint32 ts_usec;        /* timestamp microseconds */
        guint32 incl_len;       /* number of octets of packet saved in file */
        guint32 orig_len;       /* actual length of packet */
} pcaprec_hdr_t;
"""

LINKTYPE_ETHERNET = 1
LINKTYPE_RAW = 101


class PcapdService:
    SERVICE_NAME = 'com.apple.pcapd'

    def __init__(self, lockdown: LockdownClient):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        self.service = self.lockdown.start_service(self.SERVICE_NAME)

    def watch(self, out=None):
        if out:
            out.write(struct.pack("<LHHLLLL", 0xa1b2c3d4, 2, 4, 0, 0, 65535, LINKTYPE_ETHERNET))

        while True:
            d = self.service.recv_plist()
            if not d:
                break
            hdrsize, xxx, packet_size = struct.unpack(">LBL", d[:9])
            flags1, flags2, offset_to_ip_data, zero = struct.unpack(">LLLL", d[9:0x19])

            assert hdrsize >= 0x19
            interfacetype = d[0x19:hdrsize].strip(b"\x00")
            packet = d[hdrsize:]
            logging.info(packet)
            assert packet_size == len(packet)

            if offset_to_ip_data == 0:
                # add fake ethernet header for pdp packets
                packet = b"\xBE\xEF" * 6 + b"\x08\x00" + packet

            if out:
                t = time.time()
                # TODO: check milliseconds conversion
                pkthdr = struct.pack('<LLLL', int(t), int(t * 1000000 % 1000000), len(packet), len(packet))
                data = pkthdr + packet
                out.write(data)



# -*- coding: utf-8 -*-


import unittest
import os
import sys

test_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.dirname(test_dir))

def main():
    runner = unittest.TextTestRunner(verbosity=1 + sys.argv.count('-v'))
    suite = unittest.TestLoader().discover(test_dir,  pattern='*test.py')
    raise SystemExit(not runner.run(suite).wasSuccessful())


if __name__ == '__main__':
    main()


#!/usr/bin/python
import plistlib
import select
import socket
import struct
import sys


class MuxError(Exception):
    pass


class MuxVersionError(MuxError):
    pass


class SafeStreamSocket:
    def __init__(self, address, family):
        self.sock = socket.socket(family, socket.SOCK_STREAM)
        self.sock.connect(address)

    def send(self, msg):
        totalsent = 0
        while totalsent < len(msg):
            sent = self.sock.send(msg[totalsent:])
            if sent == 0:
                raise MuxError("socket connection broken")
            totalsent = totalsent + sent

    def recv(self, size):
        msg = b''
        while len(msg) < size:
            chunk = self.sock.recv(size - len(msg))
            empty_chunk = b''
            if chunk == empty_chunk:
                raise MuxError("socket connection broken")
            msg = msg + chunk
        return msg


class MuxDevice(object):
    def __init__(self, devid, usbprod, serial, location):
        self.devid = devid
        self.usbprod = usbprod
        self.serial = serial
        self.location = location

    def __str__(self):
        return "<MuxDevice: ID %d ProdID 0x%04x Serial '%s' Location 0x%x>" % (
            self.devid, self.usbprod, self.serial, self.location)


class BinaryProtocol(object):
    TYPE_RESULT = 1
    TYPE_CONNECT = 2
    TYPE_LISTEN = 3
    TYPE_DEVICE_ADD = 4
    TYPE_DEVICE_REMOVE = 5
    VERSION = 0

    def __init__(self, socket):
        self.socket = socket
        self.connected = False

    def _pack(self, req, payload):
        if req == self.TYPE_CONNECT:
            connect_data = b"\x00\x00"
            return struct.pack("IH", payload['DeviceID'], payload['PortNumber']) + connect_data
        elif req == self.TYPE_LISTEN:
            return b""
        else:
            raise ValueError("Invalid outgoing request type %d" % req)

    def _unpack(self, resp, payload):
        if resp == self.TYPE_RESULT:
            return {'Number': struct.unpack("I", payload)[0]}
        elif resp == self.TYPE_DEVICE_ADD:
            devid, usbpid, serial, pad, location = struct.unpack("IH256sHI", payload)
            serial = serial.split(b"\0")[0]
            return {'DeviceID': devid,
                    'Properties': {'LocationID': location, 'SerialNumber': serial, 'ProductID': usbpid}}
        elif resp == self.TYPE_DEVICE_REMOVE:
            devid = struct.unpack("I", payload)[0]
            return {'DeviceID': devid}
        else:
            raise MuxError("Invalid incoming response type %d" % resp)

    def send_packet(self, req, tag, payload=None):
        if payload is None:
            payload = {}
        payload = self._pack(req, payload)
        if self.connected:
            raise MuxError("Mux is connected, cannot issue control packets")
        length = 16 + len(payload)
        data = struct.pack("IIII", length, self.VERSION, req, tag) + payload
        self.socket.send(data)

    def get_packet(self):
        if self.connected:
            raise MuxError("Mux is connected, cannot issue control packets")
        dlen = self.socket.recv(4)
        dlen = struct.unpack("I", dlen)[0]
        body = self.socket.recv(dlen - 4)
        version, resp, tag = struct.unpack("III", body[:0xc])
        if version != self.VERSION:
            raise MuxVersionError("Version mismatch: expected %d, got %d" % (self.VERSION, version))
        payload = self._unpack(resp, body[0xc:])
        return (resp, tag, payload)


class PlistProtocol(BinaryProtocol):
    TYPE_RESULT = "Result"
    TYPE_CONNECT = "Connect"
    TYPE_LISTEN = "Listen"
    TYPE_DEVICE_ADD = "Attached"
    TYPE_DEVICE_REMOVE = "Detached"  # ???
    TYPE_PLIST = 8
    VERSION = 1

    def __init__(self, socket):
        BinaryProtocol.__init__(self, socket)

    def _pack(self, req, payload):
        return payload

    def _unpack(self, resp, payload):
        return payload

    def send_packet(self, req, tag, payload=None):
        if payload is None:
            payload = {}
        payload['ClientVersionString'] = 'qt4i-usbmuxd'
        if isinstance(req, int):
            req = [self.TYPE_CONNECT, self.TYPE_LISTEN][req - 2]
        payload['MessageType'] = req
        payload['ProgName'] = 'tcprelay'
        wrapped_payload = plistlib.dumps(payload)
        BinaryProtocol.send_packet(self, self.TYPE_PLIST, tag, wrapped_payload)

    def get_packet(self):
        resp, tag, payload = BinaryProtocol.get_packet(self)
        if resp != self.TYPE_PLIST:
            raise MuxError("Received non-plist type %d" % resp)
        payload = plistlib.loads(payload)
        return payload.get('MessageType', ''), tag, payload


class MuxConnection(object):
    def __init__(self, socketpath, protoclass):
        self.socketpath = socketpath
        if sys.platform in ['win32', 'cygwin']:
            family = socket.AF_INET
            address = ('127.0.0.1', 27015)
        else:
            family = socket.AF_UNIX
            address = self.socketpath
        self.socket = SafeStreamSocket(address, family)
        self.proto = protoclass(self.socket)
        self.pkttag = 1
        self.devices = []

    def _getreply(self):
        while True:
            resp, tag, data = self.proto.get_packet()
            if resp == self.proto.TYPE_RESULT:
                return tag, data
            else:
                raise MuxError("Invalid packet type received: %d" % resp)

    def _processpacket(self):
        resp, tag, data = self.proto.get_packet()
        if resp == self.proto.TYPE_DEVICE_ADD:
            self.devices.append(
                MuxDevice(data['DeviceID'], data['Properties']['ProductID'], data['Properties']['SerialNumber'],
                          data['Properties']['LocationID']))
        elif resp == self.proto.TYPE_DEVICE_REMOVE:
            for dev in self.devices:
                if dev.devid == data['DeviceID']:
                    self.devices.remove(dev)
        elif resp == self.proto.TYPE_RESULT:
            raise MuxError("Unexpected result: %d" % resp)
        else:
            raise MuxError("Invalid packet type received: %d" % resp)

    def _exchange(self, req, payload=None):
        if payload is None:
            payload = {}
        mytag = self.pkttag
        self.pkttag += 1
        self.proto.send_packet(req, mytag, payload)
        recvtag, data = self._getreply()
        if recvtag != mytag:
            raise MuxError("Reply tag mismatch: expected %d, got %d" % (mytag, recvtag))
        return data['Number']

    def listen(self):
        ret = self._exchange(self.proto.TYPE_LISTEN)
        if ret != 0:
            raise MuxError("Listen failed: error %d" % ret)

    def process(self, timeout=None):
        if self.proto.connected:
            raise MuxError("Socket is connected, cannot process listener events")
        rlo, wlo, xlo = select.select([self.socket.sock], [], [self.socket.sock], timeout)
        if xlo:
            self.socket.sock.close()
            raise MuxError("Exception in listener socket")
        if rlo:
            self._processpacket()

    def connect(self, device, port):
        ret = self._exchange(self.proto.TYPE_CONNECT,
                             {'DeviceID': device.devid, 'PortNumber': ((port << 8) & 0xFF00) | (port >> 8)})
        if ret != 0:
            raise MuxError("Connect failed: error %d" % ret)
        self.proto.connected = True
        return self.socket.sock

    def close(self):
        self.socket.sock.close()


class USBMux(object):
    def __init__(self, socket_path=None):
        if socket_path is None:
            if sys.platform == 'darwin':
                socket_path = '/var/run/usbmuxd'
            else:
                socket_path = '/var/run/usbmuxd'
        self.socket_path = socket_path
        self.listener = MuxConnection(socket_path, BinaryProtocol)
        try:
            self.listener.listen()
            self.version = 0
            self.protoclass = BinaryProtocol
        except MuxVersionError:
            self.listener = MuxConnection(socket_path, PlistProtocol)
            self.listener.listen()
            self.protoclass = PlistProtocol
            self.version = 1
        self.devices = self.listener.devices

    def process(self, timeout=10.0):
        self.listener.process(timeout)

    def connect(self, device, port):
        connector = MuxConnection(self.socket_path, self.protoclass)
        return connector.connect(device, port)


class UsbmuxdClient(MuxConnection):
    def __init__(self):
        super(UsbmuxdClient, self).__init__("/var/run/usbmuxd", PlistProtocol)

    def get_pair_record(self, udid):
        tag = self.pkttag
        self.pkttag += 1
        payload = {"PairRecordID": udid}
        self.proto.send_packet("ReadPairRecord", tag, payload)
        _, recvtag, data = self.proto.get_packet()
        if recvtag != tag:
            raise MuxError("Reply tag mismatch: expected %d, got %d" % (tag, recvtag))
        pair_record = data['PairRecordData']
        pair_record = plistlib.loads(pair_record)
        return pair_record


if __name__ == "__main__":
    mux = USBMux()
    print("Waiting for devices...")
    if not mux.devices:
        mux.process(0.1)
    while True:
        for dev in mux.devices:
            print("Device:", dev)
        mux.process()



#!/usr/bin/env python3

import plistlib
import logging
import struct

from construct import Struct, Bytes, Int32ul, CString, Timestamp, Optional, Enum, Byte, Probe

from pymobiledevice3.lockdown import LockdownClient

CHUNK_SIZE = 4096
TIME_FORMAT = '%H:%M:%S'
SYSLOG_LINE_SPLITTER = '\n\x00'

syslog_t = Struct(
    Bytes(9),
    'pid' / Int32ul,
    Bytes(42),
    'timestamp' / Timestamp(Int32ul, 1., 1970),
    Bytes(9),
    'level' / Enum(Byte, Notice=0, Error=0x10, Fault=0x11),
    Bytes(1),
    Bytes(60),
    'filename' / CString('utf8'),
    'image_name' / CString('utf8'),
    'message' / CString('utf8'),
    'label' / Optional(Struct(
        'bundle_id' / CString('utf8'),
        'identifier' / CString('utf8')
    )),
)


class OsTraceService(object):
    SERVICE_NAME = 'com.apple.os_trace_relay'

    def __init__(self, lockdown: LockdownClient):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        self.c = self.lockdown.start_service(self.SERVICE_NAME)

    def get_pid_list(self):
        self.c.send_plist({'Request': 'PidList'})
        response = self.c.recv_raw()
        return plistlib.loads(response[1:])

    def create_archive(self) -> tuple:
        self.c.send_plist({'Request': 'CreateArchive'})
        response = self.c.recv_raw()
        length = response[0]
        return plistlib.loads(response[1:length + 1]), response[length + 1:]

    def syslog(self, pid=-1):
        self.c.send_plist({'Request': 'StartActivity', 'Pid': pid})

        length_length, = struct.unpack('<I', self.c.recv_exact(4))
        length = int(self.c.recv_exact(length_length)[::-1].hex(), 16)
        response = plistlib.loads(self.c.recv_exact(length))

        if response['Status'] != 'RequestSuccessful':
            raise Exception(f'got invalid response: {response}')

        while True:
            assert b'\x02' == self.c.recv_exact(1)
            length, = struct.unpack('<I', self.c.recv_exact(4))
            line = self.c.recv_exact(length)
            entry = syslog_t.parse(line)
            yield entry



import plistlib
import os
import datetime
import logging
import codecs

from six import PY3

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.afc import AFCClient

MOBILEBACKUP_E_SUCCESS = 0
MOBILEBACKUP_E_INVALID_ARG = -1
MOBILEBACKUP_E_PLIST_ERROR = -2
MOBILEBACKUP_E_MUX_ERROR = -3
MOBILEBACKUP_E_BAD_VERSION = -4
MOBILEBACKUP_E_REPLY_NOT_OK = -5
MOBILEBACKUP_E_UNKNOWN_ERROR = -256

DEVICE_LINK_FILE_STATUS_NONE = 0
DEVICE_LINK_FILE_STATUS_HUNK = 1
DEVICE_LINK_FILE_STATUS_LAST_HUNK = 2


class DeviceVersionNotSupported(Exception):
    def __str__(self):
        return "Device version not supported, please use mobilebackup2"


class MobileBackup(object):
    def __init__(self, lockdown=None, udid=None, logger=None):
        self.logger = logger or logging.getLogger(__name__)
        self.lockdown = lockdown if lockdown else LockdownClient(udid=udid)
        ProductVersion = self.lockdown.get_value("", "ProductVersion")
        if ProductVersion[0] >= "5":
            raise DeviceVersionNotSupported
        self.start()

    def start(self):
        self.service = self.lockdown.start_service("com.apple.mobilebackup")
        self.udid = self.lockdown.udid
        DLMessageVersionExchange = self.service.recv_plist()
        version_major = DLMessageVersionExchange[1]
        self.service.send_plist(["DLMessageVersionExchange", "DLVersionsOk", version_major])
        DLMessageDeviceReady = self.service.recv_plist()
        if DLMessageDeviceReady and DLMessageDeviceReady[0] == "DLMessageDeviceReady":
            self.logger.info("Got DLMessageDeviceReady")

    def check_filename(self, name):
        if PY3 and not isinstance(name, str):
            name = codecs.decode(name)
        if "../" in name:
            raise Exception("HAX, sneaky dots in path %s" % name)
        if not name.startswith(self.backupPath):
            if name.startswith(self.udid):
                name = os.path.join(self.backupPath, name)
                return name
            name = os.path.join(self.backupPath, self.udid, name)
            return name
        return name

    def read_file(self, filename):
        filename = self.check_filename(filename)
        if os.path.isfile(filename):
            with open(filename, 'rb') as f:
                data = f.read()
                f.close()
                return data
        return None

    def write_file(self, filename, data):
        filename = self.check_filename(filename)
        with open(filename, 'wb') as f:
            f.write(data)
            f.close()

    def create_info_plist(self):
        root_node = self.lockdown.all_values
        info = {"BuildVersion": root_node.get("BuildVersion") or "", "DeviceName": root_node.get("DeviceName") or "",
                "Display Name": root_node.get("DeviceName") or "", "GUID": "---",
                "ProductType": root_node.get("ProductType") or "",
                "ProductVersion": root_node.get("ProductVersion") or "",
                "Serial Number": root_node.get("SerialNumber") or "", "Unique Identifier": self.udid.upper(),
                "Target Identifier": self.udid, "Target Type": "Device", "iTunes Version": "10.0.1",
                "ICCID": root_node.get("IntegratedCircuitCardIdentity") or "",
                "IMEI": root_node.get("InternationalMobileEquipmentIdentity") or "",
                "Last Backup Date": datetime.datetime.now()}

        afc = AFCClient(self.lockdown)
        iTunesFilesDict = {}
        iTunesFiles = afc.read_directory("/iTunes_Control/iTunes/")

        for i in iTunesFiles:
            data = afc.get_file_contents("/iTunes_Control/iTunes/" + i)
            if data:
                iTunesFilesDict[i] = plistlib.Data(data)
        info["iTunesFiles"] = iTunesFilesDict

        iBooksData2 = afc.get_file_contents("/Books/iBooksData2.plist")
        if iBooksData2:
            info["iBooks Data 2"] = plistlib.Data(iBooksData2)

        info["iTunes Settings"] = self.lockdown.get_value("com.apple.iTunes")
        self.logger.info("Creating: %s", os.path.join(self.udid, "Info.plist"))
        self.write_file(os.path.join(self.udid, "Info.plist"), plistlib.writePlistToString(info))

    def ping(self, message):
        self.service.send_plist(["DLMessagePing", message])
        res = self.service.recv_plist()
        self.logger.debug("ping response:", res)

    def device_link_service_send_process_message(self, msg):
        return self.service.send_plist(["DLMessageProcessMessage", msg])

    def device_link_service_receive_process_message(self):
        req = self.service.recv_plist()
        if req:
            assert req[0] == "DLMessageProcessMessage"
            return req[1]

    def send_file_received(self):
        return self.device_link_service_send_process_message(
            {"BackupMessageTypeKey": "kBackupMessageBackupFileReceived"})

    def request_backup(self):
        req = {"BackupComputerBasePathKey": "/",
               "BackupMessageTypeKey": "BackupMessageBackupRequest",
               "BackupProtocolVersion": "1.6"
               }
        self.create_info_plist()
        self.device_link_service_send_process_message(req)
        res = self.device_link_service_receive_process_message()
        if not res:
            return
        if res["BackupMessageTypeKey"] != "BackupMessageBackupReplyOK":
            self.logger.error(res)
            return
        self.device_link_service_send_process_message(res)

        f = None
        outpath = None
        while True:
            res = self.service.recv_plist()
            if not res or res[0] != "DLSendFile":
                if res[0] == "DLMessageProcessMessage":
                    if res[1].get("BackupMessageTypeKey") == "BackupMessageBackupFinished":
                        self.logger.info("Backup finished OK !")
                        # TODO: BackupFilesToDeleteKey
                        with open(self.check_filename("Manifest.plist"), 'wb') as f:
                            plistlib.dump(res[1]["BackupManifestKey"], f)
                break
            data = res[1].data
            info = res[2]
            if not f:
                outpath = self.check_filename(info.get("DLFileDest"))
                self.logger.debug("%s %s", info["DLFileAttributesKey"]["Filename"], info.get("DLFileDest"))
                f = open(outpath + ".mddata", "wb")
            f.write(data)
            if info.get("DLFileStatusKey") == DEVICE_LINK_FILE_STATUS_LAST_HUNK:
                self.send_file_received()
                f.close()
                if not info.get("BackupManifestKey", False):
                    with open(outpath + ".mdinfo", 'wb') as f:
                        plistlib.dump(info.get("BackupFileInfo"), f)
                f = None


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    mb = MobileBackup()
    mb.request_backup()



import posixpath
import logging
import os

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.afc import AFCClient

client_options = {
    "SkipUninstall": False,
    "ApplicationSINF": False,
    "iTunesMetadata": False,
    "ReturnAttributes": False
}


class InstallationProxyService(object):
    SERVICE_NAME = 'com.apple.mobile.installation_proxy'

    def __init__(self, lockdown: LockdownClient):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        self.service = self.lockdown.start_service(self.SERVICE_NAME)

    def watch_completion(self, handler=None, *args):
        while True:
            response = self.service.recv_plist()
            if not response:
                break
            error = response.get('Error')
            if error:
                raise IOError(f'{error}: {response.get("ErrorDescription")}')
            completion = response.get("PercentComplete")
            if completion:
                if handler:
                    self.logger.debug("calling handler")
                    handler(completion, *args)
                self.logger.info("%s %% Complete", response.get("PercentComplete"))
            if response.get("Status") == "Complete":
                return response.get("Status")
        return "Error"

    def send_cmd_for_bid(self, bid, cmd="Archive", options=None, handler=None, *args):
        cmd = {"Command": cmd,
               "ApplicationIdentifier": bid}
        if options:
            cmd.update({"ClientOptions": options})
        self.service.send_plist(cmd)
        self.logger.info("%s : %s\n", cmd, self.watch_completion(handler, *args))

    def uninstall(self, bid, options=None, handler=None, *args):
        return self.send_cmd_for_bid(bid, "Uninstall", options, handler, args)

    def install_from_local(self, ipa_path, cmd="Install", options=None, handler=None, *args):
        if options is None:
            options = {}
        remote_path = posixpath.join('/', os.path.basename(ipa_path))
        afc = AFCClient(self.lockdown)
        afc.set_file_contents(remote_path, open(ipa_path, "rb").read())
        cmd = {"Command": cmd,
               "ClientOptions": options,
               "PackagePath": remote_path}
        self.service.send_plist(cmd)
        self.watch_completion(handler, args)

    def install(self, ipa_path, options=None, handler=None, *args):
        if options is None:
            options = {}
        return self.install_from_local(ipa_path, "Install", options, handler, args)

    def upgrade(self, ipa_path, options=None, handler=None, *args):
        if options is None:
            options = {}
        return self.install_from_local(ipa_path, "Upgrade", options, handler, args)

    def check_capabilities_match(self, capabilities, options=None):
        if options is None:
            options = {}
        cmd = {"Command": "CheckCapabilitiesMatch",
               "ClientOptions": options}

        if capabilities:
            cmd["Capabilities"] = capabilities

        self.service.send_plist(cmd)
        result = self.service.recv_plist().get("LookupResult")
        return result

    def browse(self, options=None, attributes=None):
        if options is None:
            options = {}
        if attributes:
            options["ReturnAttributes"] = attributes

        cmd = {"Command": "Browse",
               "ClientOptions": options}

        self.service.send_plist(cmd)

        result = []
        while True:
            z = self.service.recv_plist()
            if not z:
                break

            data = z.get("CurrentList")
            if data:
                result += data

            if z.get("Status") == "Complete":
                break

        return result

    def apps_info(self, options=None):
        if options is None:
            options = {}
        cmd = {"Command": "Lookup",
               "ClientOptions": options}

        self.service.send_plist(cmd)
        return self.service.recv_plist().get('LookupResult')

    def archive(self, bid, options=None, handler=None, *args):
        if options is None:
            options = {}
        self.send_cmd_for_bid(bid, "Archive", options, handler, args)

    def restore_archive(self, bid, options=None, handler=None, *args):
        if options is None:
            options = {}
        self.send_cmd_for_bid(bid, "Restore", options, handler, args)

    def remove_archive(self, bid, options=None, handler=None, *args):
        if options is None:
            options = {}
        self.send_cmd_for_bid(bid, "RemoveArchive", options, handler, args)

    def archives_info(self, options=None):
        if options is None:
            options = {}
        cmd = {"Command": "LookupArchive",
               "ClientOptions": options}
        return self.service.send_request(cmd).get("LookupResult")

    def search_path_for_bid(self, bid):
        path = None
        for a in self.get_apps(app_types=["User", "System"]):
            if a.get("CFBundleIdentifier") == bid:
                path = a.get("Path") + "/" + a.get("CFBundleExecutable")
        return path

    def get_apps(self, app_types=None):
        if app_types is None:
            app_types = ["User"]
        return [app for app in self.apps_info().values()
                if app.get("ApplicationType") in app_types]

    def print_apps(self, app_types=None):
        if app_types is None:
            app_types = ["User"]
        for app in self.get_apps(app_types):
            print(("%s : %s => %s" % (app.get("CFBundleDisplayName"),
                                      app.get("CFBundleIdentifier"),
                                      app.get("Path") if app.get("Path")
                                      else app.get("Container"))).encode('utf-8'))

    def get_apps_bid(self, app_types=None):
        if app_types is None:
            app_types = ["User"]
        return [app["CFBundleIdentifier"]
                for app in self.get_apps()
                if app.get("ApplicationType") in app_types]

    def close(self):
        self.service.close()



import logging

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.afc import AFCClient, AFCShell


class HouseArrestService(AFCClient):
    SERVICE_NAME = 'com.apple.mobile.house_arrest'

    def __init__(self, lockdown: LockdownClient):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        service_name = self.SERVICE_NAME
        super(HouseArrestService, self).__init__(self.lockdown, service_name)

    def send_command(self, bundle_id, cmd="VendContainer"):
        self.service.send_plist({"Command": cmd, "Identifier": bundle_id})
        res = self.service.recv_plist()
        if res.get("Error"):
            self.logger.error("%s: %s", bundle_id, res.get("Error"))
            return False
        else:
            return True

    def shell(self, application_id, cmd="VendContainer"):
        res = self.send_command(application_id, cmd)
        if res:
            AFCShell(client=self).cmdloop()



import logging

from pymobiledevice3.lockdown import LockdownClient


class MobileImageMounterService(object):
    SERVICE_NAME = 'com.apple.mobile.mobile_image_mounter'

    def __init__(self, lockdown: LockdownClient):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        self.service = self.lockdown.start_service(self.SERVICE_NAME)

    def list_images(self):
        """ Lookup mounted image by its name. """
        self.service.send_plist({'Command': 'CopyDevices'})
        return self.service.recv_plist()

    def lookup_image(self, image_type):
        """ Lookup mounted image by its name. """
        self.service.send_plist({'Command': 'LookupImage',
                                 'ImageType': image_type})

        return self.service.recv_plist()

    def umount(self, image_type, mount_path, signature):
        """ umount image. """
        self.service.send_plist({'Command': 'UnmountImage',
                                 'ImageType': image_type,
                                 'MountPath': mount_path,
                                 'ImageSignature': signature})
        return self.service.recv_plist()

    def mount(self, image_type, signature):
        """ Upload image into device. """
        self.service.send_plist({'Command': 'MountImage',
                                 'ImageType': image_type,
                                 'ImageSignature': signature})
        result = self.service.recv_plist()
        status = result.get('Status')

        if status != 'Complete':
            raise Exception(f'command MountImage failed with: {result}')

    def upload_image(self, image_type, image, signature):
        """ Upload image into device. """
        self.service.send_plist({'Command': 'ReceiveBytes',
                                 'ImageType': image_type,
                                 'ImageSize': len(image),
                                 'ImageSignature': signature})
        result = self.service.recv_plist()

        status = result.get('Status')

        if status != 'ReceiveBytesAck':
            raise Exception(f'command ReceiveBytes failed with: {result}')

        self.service.send(image)
        result = self.service.recv_plist()

        status = result.get('Status')

        if status != 'Complete':
            raise Exception(f'command ReceiveBytes failed to send bytes with: {result}')



import logging
import io
import plistlib
from datetime import datetime
from functools import partial

import IPython
from pygments import highlight, lexers, formatters
from bpylist2 import archiver
from construct import Struct, Int32ul, Int16ul, Int64ul, Const, Prefixed, GreedyBytes, this, Adapter, Select, \
    GreedyRange, Switch

from pymobiledevice3.lockdown import LockdownClient

SHELL_USAGE = '''
# This shell allows you to send messages to the DVTSecureSocketProxy and receive answers easily.
# Generally speaking, each channel represents a group of actions. 
# Calling actions is done using a selector and auxiliary (parameters).
# Receiving answers is done by getting a return value and seldom auxiliary (private / extra parameters).
# To see the available channels, type the following:
developer.channels

# In order to send messages, you need to create a channel:
channel = developer.make_channel('com.apple.instruments.server.services.deviceinfo')

# After creating the channel you can call allowed selectors:
channel.runningProcesses()

# If an answer is expected, you can receive it using the receive method:
processes = channel.receive()

# Sometimes the selector requires parameters, You can add them using MessageAux. For example lets kill a process:
channel = developer.make_channel('com.apple.instruments.server.services.processcontrol')
args = MessageAux().append_obj(80) # This will kill pid 80
channel.killPid_(args, expects_reply=False) # Killing a process doesn't require an answer.

# In some rare cases, you might want to receive the auxiliary and the selector return value.
# For that cases you can use the recv_message method.
return_value, auxiliary = developer.recv_message()
'''


class BplitAdapter(Adapter):
    def _decode(self, obj, context, path):
        return archiver.unarchive(obj)

    def _encode(self, obj, context, path):
        return archiver.archive(obj)


dtx_message_header_struct = Struct(
    'magic' / Const(0x1F3D5B79, Int32ul),
    'cb' / Int32ul,
    'fragmentId' / Int16ul,
    'fragmentCount' / Int16ul,
    'length' / Int32ul,
    'identifier' / Int32ul,
    'conversationIndex' / Int32ul,
    'channelCode' / Int32ul,
    'expectsReply' / Int32ul,
)

dtx_message_payload_header_struct = Struct(
    'flags' / Int32ul,
    'auxiliaryLength' / Int32ul,
    'totalLength' / Int64ul,
)

message_aux_t_struct = Struct(
    'magic' / Select(Const(0x1f0, Int64ul), Const(0x1df0, Int64ul)),
    'aux' / Prefixed(Int64ul, GreedyRange(Struct(
        '_empty_dictionary' / Select(Const(0xa, Int32ul), Int32ul),
        'type' / Int32ul,
        'value' / Switch(this.type, {2: BplitAdapter(Prefixed(Int32ul, GreedyBytes)), 3: Int32ul, 4: Int64ul},
                         default=GreedyBytes),
    )))
)


class MessageAux:
    def __init__(self):
        self.values = []

    def append_int(self, value: int):
        self.values.append({'type': 3, 'value': value})
        return self

    def append_long(self, value: int):
        self.values.append({'type': 4, 'value': value})
        return self

    def append_obj(self, value):
        self.values.append({'type': 2, 'value': value})
        return self

    def __bytes__(self):
        return message_aux_t_struct.build(dict(aux=self.values))


class Channel(int):
    @classmethod
    def create(cls, value: int, service):
        channel = cls(value)
        channel._service = service
        return channel

    def receive(self):
        return self._service.recv_message()[0]

    @staticmethod
    def _sanitize_name(name: str):
        """
        Sanitize python name to ObjectiveC name.
        """
        if name.startswith('_'):
            name = '_' + name[1:].replace('_', ':')
        else:
            name = name.replace('_', ':')
        return name

    def __getattr__(self, item):
        return partial(self._service.send_message, self, self._sanitize_name(item))


class DvtSecureSocketProxyService(object):
    SERVICE_NAME = 'com.apple.instruments.remoteserver.DVTSecureSocketProxy'
    INSTRUMENTS_MESSAGE_TYPE = 2
    EXPECTS_REPLY_MASK = 0x1000
    DEVICEINFO_IDENTIFIER = 'com.apple.instruments.server.services.deviceinfo'
    APP_LISTING_IDENTIFIER = 'com.apple.instruments.server.services.device.applictionListing'
    PROCESS_CONTROL_IDENTIFIER = 'com.apple.instruments.server.services.processcontrol'

    def __init__(self, lockdown: LockdownClient):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        self.c = self.lockdown.start_service(self.SERVICE_NAME)
        self.channels = {}
        self.cur_channel = 0
        self.cur_message = 0

    def shell(self):
        IPython.embed(
            header=highlight(SHELL_USAGE, lexers.PythonLexer(), formatters.TerminalTrueColorFormatter(style='native')),
            user_ns={
                'developer': self,
                'MessageAux': MessageAux,
            })

    def ls(self, path):
        channel = self.make_channel(self.DEVICEINFO_IDENTIFIER)
        args = MessageAux().append_obj(path)
        self.send_message(
            channel, 'directoryListingForPath:', args
        )
        ret, aux = self.recv_message()
        assert ret
        return ret

    def proclist(self):
        channel = self.make_channel(self.DEVICEINFO_IDENTIFIER)
        self.send_message(channel, 'runningProcesses')
        ret, aux = self.recv_message()
        assert isinstance(ret, list)
        for process in ret:
            if 'startDate' in process:
                process['startDate'] = datetime.fromtimestamp(process['startDate'])
        return ret

    def applist(self):
        channel = self.make_channel(self.APP_LISTING_IDENTIFIER)
        args = MessageAux().append_obj({}).append_obj('')
        self.send_message(channel, 'installedApplicationsMatching:registerUpdateToken:', args)
        ret, aux = self.recv_message()
        assert isinstance(ret, list)
        return ret

    def kill(self, pid):
        channel = self.make_channel(self.PROCESS_CONTROL_IDENTIFIER)
        self.send_message(channel, 'killPid:', MessageAux().append_obj(pid), False)

    def launch(self, bundle_id):
        channel = self.make_channel(self.PROCESS_CONTROL_IDENTIFIER)
        args = MessageAux().append_obj('').append_obj(bundle_id).append_obj({}).append_obj([]).append_obj({
            'StartSuspendedKey': 0,
            'KillExisting': 1,
        })
        self.send_message(
            channel, 'launchSuspendedProcessWithDevicePath:bundleIdentifier:environment:arguments:options:', args
        )
        ret, aux = self.recv_message()
        assert ret
        return ret

    def system_information(self):
        return self._request_information('systemInformation')

    def hardware_information(self):
        return self._request_information('hardwareInformation')

    def network_information(self):
        return self._request_information('networkInformation')

    def perform_handshake(self):
        args = MessageAux()
        args.append_obj({'com.apple.private.DTXBlockCompression': 2, 'com.apple.private.DTXConnection': 1})
        self.send_message(0, '_notifyOfPublishedCapabilities:', args, expects_reply=False)
        ret, aux = self.recv_message()
        if ret != '_notifyOfPublishedCapabilities:':
            raise ValueError('Invalid answer')
        if not len(aux[0]):
            raise ValueError('Invalid answer')
        self.channels = aux[0].value

    def make_channel(self, identifier):
        assert identifier in self.channels
        self.cur_channel += 1
        code = self.cur_channel
        args = MessageAux().append_int(code).append_obj(identifier)
        self.send_message(0, '_requestChannelWithCode:identifier:', args)
        ret, aux = self.recv_message()
        assert ret is None
        assert code > 0
        return Channel.create(code, self)

    def send_message(self, channel: int, selector: str = None, args: MessageAux = None, expects_reply: bool = True):
        self.cur_message += 1

        aux = bytes(args) if args is not None else b''
        sel = archiver.archive(selector) if selector is not None else b''
        flags = self.INSTRUMENTS_MESSAGE_TYPE
        if expects_reply:
            flags |= self.EXPECTS_REPLY_MASK
        pheader = dtx_message_payload_header_struct.build(dict(flags=flags, auxiliaryLength=len(aux),
                                                               totalLength=len(aux) + len(sel)))
        mheader = dtx_message_header_struct.build(dict(
            cb=dtx_message_header_struct.sizeof(),
            fragmentId=0,
            fragmentCount=1,
            length=dtx_message_payload_header_struct.sizeof() + len(aux) + len(sel),
            identifier=self.cur_message,
            conversationIndex=0,
            channelCode=channel,
            expectsReply=int(expects_reply)
        ))
        msg = mheader + pheader + aux + sel
        self.c.send(msg)

    def recv_message(self):
        packet_stream = self._recv_packet_fragments()
        pheader = dtx_message_payload_header_struct.parse_stream(packet_stream)

        compression = (pheader.flags & 0xFF000) >> 12
        if compression:
            raise NotImplementedError('Compressed')

        if pheader.auxiliaryLength:
            aux = message_aux_t_struct.parse_stream(packet_stream).aux
        else:
            aux = None
        obj_size = pheader.totalLength - pheader.auxiliaryLength
        data = packet_stream.read(obj_size)
        try:
            ret = archiver.unarchive(data) if obj_size else None
        except archiver.MissingClassMapping as e:
            print(plistlib.loads(data))
            raise e
        return ret, aux

    def _request_information(self, selector_name):
        channel = self.make_channel(self.DEVICEINFO_IDENTIFIER)
        self.send_message(channel, selector_name)
        ret, aux = self.recv_message()
        assert ret
        return ret

    def _recv_packet_fragments(self):
        packet_data = b''
        while True:
            data = self.c.recv_exact(dtx_message_header_struct.sizeof())
            mheader = dtx_message_header_struct.parse(data)
            if not mheader.conversationIndex:
                if mheader.identifier > self.cur_message:
                    self.cur_message = mheader.identifier
            if mheader.fragmentCount > 1 and mheader.fragmentId == 0:
                # when reading multiple message fragments, the first fragment contains only a message header
                continue
            packet_data += self.c.recv_exact(mheader.length)
            if mheader.fragmentId == mheader.fragmentCount - 1:
                break
        return io.BytesIO(packet_data)

    def __enter__(self):
        self.perform_handshake()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



import plistlib
import logging

from pymobiledevice3.lockdown import LockdownClient


class MobileConfigService(object):
    SERVICE_NAME = 'com.apple.mobile.MCInstall'

    def __init__(self, lockdown: LockdownClient):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        self.service = lockdown.start_service(self.SERVICE_NAME)

    def get_profile_list(self):
        self.service.send_plist({"RequestType": "GetProfileList"})
        response = self.service.recv_plist()
        if response.get("Status", None) != "Acknowledged":
            raise Exception(f'invalid response {response}')
        return response

    def install_profile(self, payload):
        self.service.send_plist({"RequestType": "InstallProfile", "Payload": plistlib.Data(payload)})
        return self.service.recv_plist()

    def remove_profile(self, ident):
        profiles = self.get_profile_list()
        if not profiles:
            return
        if ident not in profiles["ProfileMetadata"]:
            self.logger.info("Trying to remove not installed profile %s", ident)
            return
        meta = profiles["ProfileMetadata"][ident]
        data = plistlib.dumps({"PayloadType": "Configuration",
                               "PayloadIdentifier": ident,
                               "PayloadUUID": meta["PayloadUUID"],
                               "PayloadVersion": meta["PayloadVersion"]
                               })
        self.service.send_plist({"RequestType": "RemoveProfile", "ProfileIdentifier": plistlib.Data(data)})
        return self.service.recv_plist()



from re import sub
import plistlib
import logging
import struct
import ssl

import IPython
from pygments import highlight, lexers, formatters

from pymobiledevice3 import usbmux


class ConnectionFailedException(Exception):
    pass


SHELL_USAGE = """
# This shell allows you to communicate directly with every service layer behind the lockdownd daemon.

# For example, you can do the following:
client.send_plist({"Command": "DoSomething"})

# and view the reply
print(client.recv_plist())

# or just send raw message
client.send(b"hello")

# and view the result
print(client.recv_exact(20))
"""


class ServiceConnection(object):
    def __init__(self, socket):
        self.logger = logging.getLogger(__name__)
        self.socket = socket

    @staticmethod
    def create(udid, port):
        mux = usbmux.USBMux()
        mux.process(1.0)
        target_device = None

        while target_device is None:
            mux.process(1.0)
            for connected_device in mux.devices:
                if connected_device.serial == udid:
                    target_device = connected_device
                    break
        try:
            socket = mux.connect(target_device, port)
        except:
            raise ConnectionFailedException(f'Connection to device port {port} failed')

        return ServiceConnection(socket)

    def setblocking(self, blocking: bool):
        self.socket.setblocking(blocking)

    def close(self):
        self.socket.close()

    def recv(self, length=4096):
        data = self.socket.recv(length)
        return data

    def send(self, data):
        self.socket.sendall(data)

    def send_request(self, data):
        self.send_plist(data)
        return self.recv_plist()

    def recv_exact(self, l):
        data = b""
        while l > 0:
            d = self.recv(l)
            if not d or len(d) == 0:
                break
            data += d
            l -= len(d)
        return data

    def recv_raw(self):
        l = self.recv_exact(4)
        if not l or len(l) != 4:
            return
        l = struct.unpack(">L", l)[0]
        return self.recv_exact(l)

    def send_raw(self, data):
        if isinstance(data, str):
            data = data.encode()
        hdr = struct.pack(">L", len(data))
        msg = b"".join([hdr, data])
        return self.send(msg)

    def recv_plist(self):
        payload = self.recv_raw()
        if not payload:
            return
        bplist_header = b'bplist00'
        xml_header = b'<?xml'
        if payload.startswith(bplist_header):
            return plistlib.loads(payload)
        elif payload.startswith(xml_header):
            # HAX lockdown HardwarePlatform with null bytes
            payload = sub('[^\w<>\/ \-_0-9\"\'\\=\.\?\!\+]+', '', payload.decode('utf-8')).encode('utf-8')
            return plistlib.loads(payload)
        else:
            raise Exception(f'recv_plist invalid data: {payload[:100].hex()}')

    def send_plist(self, d):
        payload = plistlib.dumps(d)
        l = struct.pack(">L", len(payload))
        return self.send(l + payload)

    def ssl_start(self, keyfile, certfile):
        self.socket = ssl.wrap_socket(self.socket, keyfile, certfile, ssl_version=ssl.PROTOCOL_TLSv1)

    def shell(self):
        IPython.embed(
            header=highlight(SHELL_USAGE, lexers.PythonLexer(), formatters.TerminalTrueColorFormatter(style='native')),
            user_ns={
                'client': self,
            })



#!/usr/bin/env python3

from pprint import pprint
from cmd import Cmd
import posixpath
import plistlib
import logging
import struct
import os

from construct import Struct, Const, Int64ul, Container, Switch, Enum, Bytes, Tell, this
import hexdump

from pymobiledevice3.lockdown import LockdownClient

MAXIMUM_READ_SIZE = 1 << 16
MODE_MASK = 0o0000777

afc_opcode_t = Enum(Int64ul,
                    AFC_OP_STATUS=0x00000001,
                    AFC_OP_DATA=0x00000002,  # Data */
                    AFC_OP_READ_DIR=0x00000003,  # ReadDir */
                    AFC_OP_READ_FILE=0x00000004,  # ReadFile */
                    AFC_OP_WRITE_FILE=0x00000005,  # WriteFile */
                    AFC_OP_WRITE_PART=0x00000006,  # WritePart */
                    AFC_OP_TRUNCATE=0x00000007,  # TruncateFile */
                    AFC_OP_REMOVE_PATH=0x00000008,  # RemovePath */
                    AFC_OP_MAKE_DIR=0x00000009,  # MakeDir */
                    AFC_OP_GET_FILE_INFO=0x0000000a,  # GetFileInfo */
                    AFC_OP_GET_DEVINFO=0x0000000b,  # GetDeviceInfo */
                    AFC_OP_WRITE_FILE_ATOM=0x0000000c,  # WriteFileAtomic (tmp file+rename) */
                    AFC_OP_FILE_OPEN=0x0000000d,  # FileRefOpen */
                    AFC_OP_FILE_OPEN_RES=0x0000000e,  # FileRefOpenResult */
                    AFC_OP_READ=0x0000000f,  # FileRefRead */
                    AFC_OP_WRITE=0x00000010,  # FileRefWrite */
                    AFC_OP_FILE_SEEK=0x00000011,  # FileRefSeek */
                    AFC_OP_FILE_TELL=0x00000012,  # FileRefTell */
                    AFC_OP_FILE_TELL_RES=0x00000013,  # FileRefTellResult */
                    AFC_OP_FILE_CLOSE=0x00000014,  # FileRefClose */
                    AFC_OP_FILE_SET_SIZE=0x00000015,  # FileRefSetFileSize (ftruncate) */
                    AFC_OP_GET_CON_INFO=0x00000016,  # GetConnectionInfo */
                    AFC_OP_SET_CON_OPTIONS=0x00000017,  # SetConnectionOptions */
                    AFC_OP_RENAME_PATH=0x00000018,  # RenamePath */
                    AFC_OP_SET_FS_BS=0x00000019,  # SetFSBlockSize (0x800000) */
                    AFC_OP_SET_SOCKET_BS=0x0000001A,  # SetSocketBlockSize (0x800000) */
                    AFC_OP_FILE_LOCK=0x0000001B,  # FileRefLock */
                    AFC_OP_MAKE_LINK=0x0000001C,  # MakeLink */
                    AFC_OP_SET_FILE_TIME=0x0000001E,  # set st_mtime */
                    )

afc_error_t = Enum(Int64ul,
                   AFC_E_SUCCESS=0,
                   AFC_E_UNKNOWN_ERROR=1,
                   AFC_E_OP_HEADER_INVALID=2,
                   AFC_E_NO_RESOURCES=3,
                   AFC_E_READ_ERROR=4,
                   AFC_E_WRITE_ERROR=5,
                   AFC_E_UNKNOWN_PACKET_TYPE=6,
                   AFC_E_INVALID_ARG=7,
                   AFC_E_OBJECT_NOT_FOUND=8,
                   AFC_E_OBJECT_IS_DIR=9,
                   AFC_E_PERM_DENIED=10,
                   AFC_E_SERVICE_NOT_CONNECTED=11,
                   AFC_E_OP_TIMEOUT=12,
                   AFC_E_TOO_MUCH_DATA=13,
                   AFC_E_END_OF_DATA=14,
                   AFC_E_OP_NOT_SUPPORTED=15,
                   AFC_E_OBJECT_EXISTS=16,
                   AFC_E_OBJECT_BUSY=17,
                   AFC_E_NO_SPACE_LEFT=18,
                   AFC_E_OP_WOULD_BLOCK=19,
                   AFC_E_IO_ERROR=20,
                   AFC_E_OP_INTERRUPTED=21,
                   AFC_E_OP_IN_PROGRESS=22,
                   AFC_E_INTERNAL_ERROR=23,
                   AFC_E_MUX_ERROR=30,
                   AFC_E_NO_MEM=31,
                   AFC_E_NOT_ENOUGH_DATA=32,
                   AFC_E_DIR_NOT_EMPTY=33,
                   )

AFC_FOPEN_RDONLY = 0x00000001  # /**< r   O_RDONLY */
AFC_FOPEN_RW = 0x00000002  # /**< r+  O_RDWR   | O_CREAT */
AFC_FOPEN_WRONLY = 0x00000003  # /**< w   O_WRONLY | O_CREAT  | O_TRUNC */
AFC_FOPEN_WR = 0x00000004  # /**< w+  O_RDWR   | O_CREAT  | O_TRUNC */
AFC_FOPEN_APPEND = 0x00000005  # /**< a   O_WRONLY | O_APPEND | O_CREAT */
AFC_FOPEN_RDAPPEND = 0x00000006  # /**< a+  O_RDWR   | O_APPEND | O_CREAT */

AFC_HARDLINK = 1
AFC_SYMLINK = 2

AFC_LOCK_SH = 1 | 4  # /**< shared lock */
AFC_LOCK_EX = 2 | 4  # /**< exclusive lock */
AFC_LOCK_UN = 8 | 4  # /**< unlock */

# not really necessary
MAXIMUM_WRITE_SIZE = 1 << 32

AFCMAGIC = b"CFA6LPAA"

AFCPacket = Struct(
    'magic' / Const(AFCMAGIC),
    'entire_length' / Int64ul,
    'this_length' / Int64ul,
    'packet_num' / Int64ul,
    'operation' / afc_opcode_t,
    '_data_offset' / Tell,
    # 'data' / Bytes(this.entire_length - this._data_offset),
)


def list_to_dict(d):
    d = d.decode('utf-8')
    t = d.split("\x00")
    t = t[:-1]

    assert len(t) % 2 == 0
    res = {}
    for i in range(int(len(t) / 2)):
        res[t[i * 2]] = t[i * 2 + 1]
    return res


class AFCClient(object):
    def __init__(self, lockdown=None, service_name="com.apple.afc", service=None, udid=None, logger=None):
        self.logger = logger or logging.getLogger(__name__)
        self.serviceName = service_name
        self.lockdown = lockdown if lockdown else LockdownClient(udid=udid)
        self.service = service if service else self.lockdown.start_service(self.serviceName)
        self.packet_num = 0

    def _dispatch_packet(self, operation, data, this_length=0):
        afcpack = Container(magic=AFCMAGIC,
                            entire_length=AFCPacket.sizeof() + len(data),
                            this_length=AFCPacket.sizeof() + len(data),
                            packet_num=self.packet_num,
                            operation=operation)
        if this_length:
            afcpack.this_length = this_length
        header = AFCPacket.build(afcpack)
        self.packet_num += 1
        self.service.send(header + data)

    def _receive_data(self):
        res = self.service.recv_exact(AFCPacket.sizeof())
        status = afc_error_t.AFC_E_SUCCESS
        data = ""
        if res:
            res = AFCPacket.parse(res)
            assert res["entire_length"] >= AFCPacket.sizeof()
            length = res["entire_length"] - AFCPacket.sizeof()
            data = self.service.recv_exact(length)
            if res.operation == afc_opcode_t.AFC_OP_STATUS:
                if length != 8:
                    self.logger.error("Status length != 8")
                status = afc_error_t.parse(data)
            elif res.operation != afc_opcode_t.AFC_OP_DATA:
                pass  # print "error ?", res
        return status, data

    def _do_operation(self, opcode, data: bytes = b""):
        self._dispatch_packet(opcode, data)
        status, data = self._receive_data()

        if status != afc_error_t.AFC_E_SUCCESS:
            raise Exception(f'opcode: {opcode} failed with status: {status}')

        return data

    def get_device_info(self):
        return list_to_dict(self._do_operation(afc_opcode_t.AFC_OP_GET_DEVINFO))

    def read_directory(self, dirname):
        data = self._do_operation(afc_opcode_t.AFC_OP_READ_DIR, dirname.encode())
        data = data.decode('utf-8')
        return [x for x in data.split("\x00") if x != ""]

    def make_directory(self, dirname):
        return self._do_operation(afc_opcode_t.AFC_OP_MAKE_DIR, dirname.encode())

    def remove_directory(self, dirname):
        info = self.get_file_info(dirname)
        if not info or info.get("st_ifmt") != "S_IFDIR":
            self.logger.info("remove_directory: %s not S_IFDIR", dirname)
            return

        for d in self.read_directory(dirname):
            if d == "." or d == ".." or d == "":
                continue

            info = self.get_file_info(dirname + "/" + d)
            if info.get("st_ifmt") == "S_IFDIR":
                self.remove_directory(dirname + "/" + d)
            else:
                self.logger.info("%s/%s", dirname, d)
                self.file_remove(dirname + "/" + d)
        assert len(self.read_directory(dirname)) == 2  # .. et .
        return self.file_remove(dirname)

    def get_file_info(self, filename):
        return list_to_dict(self._do_operation(afc_opcode_t.AFC_OP_GET_FILE_INFO, filename.encode()))

    def make_link(self, target, linkname, type=AFC_SYMLINK):
        linkname = linkname.encode('utf-8')
        separator = b"\x00"
        return self._do_operation(afc_opcode_t.AFC_OP_MAKE_LINK,
                                  struct.pack("<Q", type) + target + separator + linkname + separator)

    def file_open(self, filename, mode=AFC_FOPEN_RDONLY):
        filename = filename.encode('utf-8')
        separator = b"\x00"
        data = self._do_operation(afc_opcode_t.AFC_OP_FILE_OPEN, struct.pack("<Q", mode) + filename + separator)
        return struct.unpack("<Q", data)[0]

    def file_close(self, handle):
        return self._do_operation(afc_opcode_t.AFC_OP_FILE_CLOSE, struct.pack("<Q", handle))

    def file_remove(self, filename):
        filename = filename.encode('utf-8')
        separator = b"\x00"
        return self._do_operation(afc_opcode_t.AFC_OP_REMOVE_PATH, filename + separator)

    def file_rename(self, old, new):
        old = old.encode('utf-8')
        new = new.encode('utf-8')
        separator = b"\x00"
        return self._do_operation(afc_opcode_t.AFC_OP_RENAME_PATH, old + separator + new + separator)

    def file_read(self, handle, sz):
        data = b""
        while sz > 0:
            if sz > MAXIMUM_READ_SIZE:
                toRead = MAXIMUM_READ_SIZE
            else:
                toRead = sz
            try:
                self._dispatch_packet(afc_opcode_t.AFC_OP_READ, struct.pack("<QQ", handle, toRead))
                status, chunk = self._receive_data()
            except:
                import traceback
                traceback.print_exc()
                self.lockdown = LockdownClient()
                self.service = self.lockdown.start_service("com.apple.afc")
                return self.file_read(handle, sz)

            if status != afc_error_t.AFC_E_SUCCESS:
                break
            sz -= toRead
            data += chunk
        return data

    def file_write(self, handle, data, chunk_size=MAXIMUM_WRITE_SIZE):
        file_handle = struct.pack("<Q", handle)
        chunks_count = len(data) // chunk_size
        b = b''
        for i in range(chunks_count):
            chunk = data[i * chunk_size:(i + 1) * chunk_size]
            self._dispatch_packet(afc_opcode_t.AFC_OP_WRITE,
                                  file_handle + chunk,
                                  this_length=48)
            b += chunk

            status, response = self._receive_data()
            if status != afc_error_t.AFC_E_SUCCESS:
                raise IOError(f'failed to write chunk: {status}')

        if len(data) % chunk_size:
            chunk = data[chunks_count * chunk_size:]
            self._dispatch_packet(afc_opcode_t.AFC_OP_WRITE,
                                  file_handle + chunk,
                                  this_length=48)

            b += chunk

            status, response = self._receive_data()
            if status != afc_error_t.AFC_E_SUCCESS:
                raise IOError(f'failed to write last chunk: {status}')

    def get_file_contents(self, filename):
        info = self.get_file_info(filename)
        if info:
            if info['st_ifmt'] == 'S_IFLNK':
                filename = info['LinkTarget']

            if info['st_ifmt'] == 'S_IFDIR':
                self.logger.info("%s is directory...", filename)
                return

            self.logger.info("Reading: %s", filename)
            h = self.file_open(filename)
            if not h:
                return
            d = self.file_read(h, int(info["st_size"]))
            self.file_close(h)
            return d
        return

    def set_file_contents(self, filename, data):
        h = self.file_open(filename, AFC_FOPEN_WR)
        self.file_write(h, data)
        self.file_close(h)

    def dir_walk(self, dirname):
        dirs = []
        files = []
        for fd in self.read_directory(dirname):
            fd = fd.decode('utf-8')
            if fd in ('.', '..', ''):
                continue
            infos = self.get_file_info(posixpath.join(dirname, fd))
            if infos and infos.get('st_ifmt') == 'S_IFDIR':
                dirs.append(fd)
            else:
                files.append(fd)

        yield dirname, dirs, files

        if dirs:
            for d in dirs:
                for walk_result in self.dir_walk(posixpath.join(dirname, d)):
                    yield walk_result


class AFCShell(Cmd):
    def __init__(self, lockdown: LockdownClient, afcname='com.apple.afc', completekey='tab'):
        Cmd.__init__(self, completekey=completekey)
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        self.afc = AFCClient(self.lockdown, service_name=afcname)
        self.curdir = '/'
        self.prompt = 'AFC$ ' + self.curdir + ' '
        self.complete_cat = self._complete
        self.complete_ls = self._complete

    def do_exit(self, p):
        return True

    def do_quit(self, p):
        return True

    def do_pwd(self, p):
        print(self.curdir)

    def do_link(self, p):
        z = p.split()
        self.afc.make_link(AFC_SYMLINK, z[0], z[1])

    def do_cd(self, p):
        if not p.startswith("/"):
            new = posixpath.join(self.curdir, p)
        else:
            new = p

        new = os.path.normpath(new).replace("\\", "/").replace("//", "/")
        if self.afc.read_directory(new):
            self.curdir = new
            self.prompt = "AFC$ %s " % new
        else:
            self.logger.error("%s does not exist", new)

    def _complete(self, text, line, begidx, endidx):
        filename = text.split("/")[-1]
        dirname = "/".join(text.split("/")[:-1])
        return [dirname + "/" + x for x in self.afc.read_directory(self.curdir + "/" + dirname) if
                x.startswith(filename)]

    def do_ls(self, p):
        dirname = posixpath.join(self.curdir, p)
        if self.curdir.endswith("/"):
            dirname = self.curdir + p
        d = self.afc.read_directory(dirname)
        if d:
            for dd in d:
                print(dd)

    def do_cat(self, p):
        data = self.afc.get_file_contents(posixpath.join(self.curdir, p))
        if data and p.endswith('.plist'):
            pprint(plistlib.loads(data))
        else:
            print(data)

    def do_rm(self, p):
        f = self.afc.get_file_info(posixpath.join(self.curdir, p))
        if f['st_ifmt'] == 'S_IFDIR':
            d = self.afc.remove_directory(posixpath.join(self.curdir, p))
        else:
            d = self.afc.file_remove(posixpath.join(self.curdir, p))

    def do_pull(self, user_args):
        args = user_args.split()
        if len(args) != 2:
            local_path = "."
            remote_path = user_args
        else:
            local_path = args[1]
            remote_path = args[0]

        remote_file_info = self.afc.get_file_info(posixpath.join(self.curdir, remote_path))
        if not remote_file_info:
            logging.error("remote file does not exist")
            return

        out_path = posixpath.join(local_path, remote_path)
        if remote_file_info['st_ifmt'] == 'S_IFDIR':
            if not os.path.isdir(out_path):
                os.makedirs(out_path, MODE_MASK)

            for d in self.afc.read_directory(remote_path):
                if d == "." or d == ".." or d == "":
                    continue
                self.do_pull(remote_path + "/" + d + " " + local_path)
        else:
            contents = self.afc.get_file_contents(posixpath.join(self.curdir, remote_path))
            out_dir = os.path.dirname(out_path)
            if not os.path.exists(out_dir):
                os.makedirs(out_dir, MODE_MASK)
            with open(out_path, 'wb') as remote_file_info:
                remote_file_info.write(contents)

    def do_push(self, p):
        src_dst = p.split()
        if len(src_dst) != 2:
            self.logger.error('USAGE: push <src> <dst>')
            return
        src = src_dst[0]
        dst = src_dst[1]

        src = os.path.expanduser(src)
        dst = os.path.expanduser(dst)

        logging.info(f'from {src} to {dst}')
        if os.path.isdir(src):
            self.afc.make_directory(os.path.join(dst))
            for x in os.listdir(src):
                if x.startswith("."):
                    continue
                path = os.path.join(src, x)
                self.do_push(path + " " + dst + "/" + path)
        else:
            data = open(src, "rb").read()
            self.afc.set_file_contents(posixpath.join(self.curdir, dst), data)

    def do_head(self, p):
        print(self.afc.get_file_contents(posixpath.join(self.curdir, p))[:32])

    def do_hexdump(self, filename):
        filename = posixpath.join(self.curdir, filename)
        print(hexdump.hexdump(self.afc.get_file_contents(filename)))

    def do_mkdir(self, p):
        print(self.afc.make_directory(p))

    def do_rmdir(self, p):
        return self.afc.remove_directory(p)

    def do_info(self, p):
        for k, v in self.afc.get_device_info().items():
            print(k, '\t:\t', v)

    def do_mv(self, p):
        t = p.split()
        return self.afc.rename_path(t[0], t[1])

    def do_stat(self, filename):
        filename = posixpath.join(self.curdir, filename)
        pprint(self.afc.get_file_info(filename))



#!/usr/bin/env python3
import plistlib
import datetime
import os

from optparse import OptionParser
from time import mktime, gmtime
from uuid import uuid4
from stat import *

from pymobiledevice3.afc import AFCClient
from pymobiledevice3.services.installation_proxy_service import InstallationProxyService
from pymobiledevice3.services.notification_proxy_service import *
from pymobiledevice3.services.springboard_services import SpringBoardServicesService
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.mobilebackup import MobileBackup

CODE_SUCCESS = 0x00
CODE_ERROR_LOCAL = 0x06
CODE_ERROR_REMOTE = 0x0b
CODE_FILE_DATA = 0x0c

ERROR_ENOENT = -6
ERROR_EEXIST = -7

LOCK_ATTEMPTS = 10


class DeviceVersionNotSupported(Exception):
    def __str__(self):
        return "Device version not supported, please use mobilebackup"


class MobileBackup2(MobileBackup):
    service = None

    def __init__(self, lockdown=None, backupPath=None, password="", udid=None, logger=None):
        self.logger = logger or logging.getLogger(__name__)
        self.backupPath = backupPath if backupPath else "backups"
        self.password = password
        self.lockdown = lockdown if lockdown else LockdownClient(udid=udid)
        if not self.lockdown:
            raise Exception("Unable to start lockdown")

        ProductVersion = self.lockdown.get_value("", "ProductVersion")
        if ProductVersion and int(ProductVersion[:ProductVersion.find('.')]) < 5:
            raise DeviceVersionNotSupported
        self.start()

    def start(self):
        self.udid = lockdown.get_value("", "UniqueDeviceID")
        self.willEncrypt = lockdown.get_value("com.apple.mobile.backup", "WillEncrypt")
        self.escrowBag = lockdown.get_value('', 'EscrowBag')
        self.afc = AFCClient(self.lockdown)  # We need this to create lock files
        self.service = self.lockdown.start_service("com.apple.mobilebackup2")
        if not self.service:
            raise Exception("MobileBackup2 init error : Could not start com.apple.mobilebackup2")

        if not os.path.isdir(self.backupPath):
            os.makedirs(self.backupPath, 0o0755)

        self.logger.info("Starting new com.apple.mobilebackup2 service with working dir: %s", self.backupPath)

        DLMessageVersionExchange = self.service.recv_plist()
        version_major = DLMessageVersionExchange[1]
        self.service.send_plist(["DLMessageVersionExchange", "DLVersionsOk", version_major])
        DLMessageDeviceReady = self.service.recv_plist()
        if DLMessageDeviceReady and DLMessageDeviceReady[0] == "DLMessageDeviceReady":
            res = self.version_exchange()
            protocol_version = res.get('ProtocolVersion')
            self.logger.info("Negotiated Protocol Version %s", protocol_version)
        else:
            raise Exception("MobileBackup2 init error %s", DLMessageDeviceReady)

    def __del__(self):
        if self.service:
            self.service.send_plist(["DLMessageDisconnect", "___EmptyParameterString___"])

    def internal_mobilebackup2_send_message(self, name, data):
        data["MessageName"] = name
        self.device_link_service_send_process_message(data)

    def internal_mobilebackup2_receive_message(self, name=None):
        res = self.device_link_service_receive_process_message()
        if res:
            if name and res["MessageName"] != name:
                self.logger.error("MessageName does not match %s %s", name, str(res))
            return res

    def version_exchange(self):
        self.internal_mobilebackup2_send_message("Hello",
                                                 {"SupportedProtocolVersions": [2.0, 2.1]})

        return self.internal_mobilebackup2_receive_message("Response")

    def mobilebackup2_send_request(self, request, target, source, options=None):
        if options is None:
            options = {}
        d = {"TargetIdentifier": target,
             "SourceIdentifier": source,
             "Options": options}

        self.internal_mobilebackup2_send_message(request, d)

    def mobilebackup2_receive_message(self):
        return self.service.recv_plist()

    def mobilebackup2_send_status_response(self, status_code,
                                           status1="___EmptyParameterString___",
                                           status2=None):
        if status2 is None:
            status2 = {}
        a = ["DLMessageStatusResponse", status_code, status1, status2]
        self.service.send_plist(a)

    def mb2_handle_free_disk_space(self, msg):
        s = os.statvfs(self.backupPath)
        freeSpace = s.f_bsize * s.f_bavail
        res = ["DLMessageStatusResponse", 0, "___EmptyParameterString___", freeSpace]
        self.service.send_plist(res)

    def mb2_multi_status_add_file_error(self, errplist, path, error_code, error_message):
        errplist[path] = {"DLFileErrorCode": error_code, "DLFileErrorString": error_message}

    def mb2_handle_copy_item(self, msg):
        src = self.check_filename(msg[1])
        dst = self.check_filename(msg[2])
        if os.path.isfile(src):
            data = self.read_file(src)
            self.write_file(dst, data)
        else:
            os.makedirs(dst)
        self.mobilebackup2_send_status_response(0)

    def mb2_handle_send_file(self, filename, errplist):
        self.service.send_raw(filename)
        if not filename.startswith(self.udid):
            filename = self.udid + "/" + filename

        data = self.read_file(self.check_filename(filename))
        if data is not None:
            self.logger.info("Sending %s to device", filename)
            msg = b"".join([(CODE_FILE_DATA).to_bytes(1, 'little'), data])
            self.service.send_raw(msg)
            self.service.send_raw(chr(CODE_SUCCESS))
        else:
            self.logger.warning("File %s requested from device not found", filename)
            self.service.send_raw(chr(CODE_ERROR_LOCAL))
            self.mb2_multi_status_add_file_error(errplist, filename,
                                                 ERROR_ENOENT, "Could not find the droid you were looking for ;)")

    def mb2_handle_send_files(self, msg):
        err_plist = {}
        for f in msg[1]:
            self.mb2_handle_send_file(f, err_plist)
        msg = b'\x00\x00\x00\x00'
        self.service.send(msg)
        if len(err_plist):
            self.mobilebackup2_send_status_response(-13, 'Multi status', err_plist)
        else:
            self.mobilebackup2_send_status_response(0)

    def mb2_handle_list_directory(self, msg):
        path = msg[1]
        self.logger.info("List directory: %s" % path)
        dirlist = {}
        if path.find("../") != -1:
            raise Exception("HAX, sneaky dots in path %s" % name)
        for root, dirs, files in os.walk(os.path.join(self.backupPath, path)):
            for fname in files:
                fpath = os.path.join(root, fname)
                finfo = {}
                st = os.stat(fpath)
                ftype = "DLFileTypeUnknown"
                if S_ISDIR(st.st_mode):
                    ftype = "DLFileTypeDirectory"
                elif S_ISREG(st.st_mode):
                    ftype = "DLFileTypeRegular"
                finfo["DLFileType"] = ftype
                finfo["DLFileSize"] = st.st_size
                finfo["DLFileModificationDate"] = st.st_mtime
                dirlist[fname] = finfo
        self.mobilebackup2_send_status_response(0, status2=dirlist);

    def mb2_handle_make_directory(self, msg):
        dirname = self.check_filename(msg[1])
        self.logger.info("Creating directory %s", dirname)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)
        self.mobilebackup2_send_status_response(0, "")

    def mb2_handle_receive_files(self, msg):
        done = 0
        while not done:
            device_filename = self.service.recv_raw()
            if device_filename == "":
                break
            backup_filename = self.service.recv_raw()
            self.logger.debug("Downloading: %s to %s", device_filename, backup_filename)
            filedata = bytearray(b"")
            last_code = 0x00
            while True:
                stuff = self.service.recv_raw()
                code = stuff[0]
                if code == CODE_FILE_DATA:
                    filedata = stuff[1:]
                elif code == CODE_SUCCESS:
                    self.write_file(self.check_filename(backup_filename), filedata)
                    break
                elif code == CODE_ERROR_REMOTE:
                    if last_code != CODE_FILE_DATA:
                        self.logger.warning("Received an error message from device: %s for:\n\t%s\n\t[%s]",
                                            ord(code), device_filename, backup_filename)
                else:
                    self.logger.warning("Unknown code: %s for:\n\t%s\n\t[%s]",
                                        code, device_filename, backup_filename)
                    self.logger.warning(msg)
                    # break
            last_code = code
        self.mobilebackup2_send_status_response(0)

    def mb2_handle_move_files(self, msg):
        self.logger.info("Moving %d files", len(msg[1]))
        for k, v in msg[1].items():
            self.logger.info("Renaming:\n\t%s \n\tto %s", self.check_filename(k), self.check_filename(v))
            os.rename(self.check_filename(k), self.check_filename(v))
        self.mobilebackup2_send_status_response(0)

    def mb2_handle_remove_files(self, msg):
        self.logger.info("Removing %d files", len(msg[1]))
        for filename in msg[1]:
            self.logger.info("Removing %s", self.check_filename(filename))
            try:
                filename = self.check_filename(filename)
                if os.path.isfile(filename):
                    os.unlink(filename)
            except Exception as e:
                self.logger.error(e)
        self.mobilebackup2_send_status_response(0)

    def work_loop(self):
        while True:
            msg = self.mobilebackup2_receive_message()
            if not msg:
                break

            assert (msg[0] in ["DLMessageDownloadFiles",
                               "DLContentsOfDirectory",
                               "DLMessageCreateDirectory",
                               "DLMessageUploadFiles",
                               "DLMessageMoveFiles", "DLMessageMoveItems",
                               "DLMessageRemoveFiles", "DLMessageRemoveItems",
                               "DLMessageCopyItem",
                               "DLMessageProcessMessage",
                               "DLMessageGetFreeDiskSpace",
                               "DLMessageDisconnect"])

            if msg[0] == "DLMessageDownloadFiles":
                self.mb2_handle_send_files(msg)
            elif msg[0] == "DLContentsOfDirectory":
                self.mb2_handle_list_directory(msg)
            elif msg[0] == "DLMessageCreateDirectory":
                self.mb2_handle_make_directory(msg)
            elif msg[0] == "DLMessageUploadFiles":
                self.mb2_handle_receive_files(msg)
            elif msg[0] in ["DLMessageMoveFiles", "DLMessageMoveItems"]:
                self.mb2_handle_move_files(msg)
            elif msg[0] in ["DLMessageRemoveFiles", "DLMessageRemoveItems"]:
                self.mb2_handle_remove_files(msg)
            elif msg[0] == "DLMessageCopyItem":
                self.mb2_handle_copy_item(msg)
            elif msg[0] == "DLMessageProcessMessage":
                errcode = msg[1].get("ErrorCode")
                if errcode == 0:
                    m = msg[1].get("MessageName")
                    if m != "Response":
                        self.logger.warning(m)
                    break
                if errcode == 1:
                    self.logger.info("Please unlock your device and retry...")
                    raise Exception("Please unlock your device and retry...")
                if errcode == 211:
                    self.logger.info("Please go to Settings->iClould->Find My iPhone and disable it")
                    raise Exception('Please go to Settings->iClould->Find My iPhone and disable it')
                if errcode == 105:
                    self.logger.info("Not enough free space on device for restore")
                    raise Exception('Not enough free space on device for restore')
                if errcode == 17:
                    self.logger.info("please press 'trust this computer' in your device")
                    raise Exception('please press \'trust this computer\' in your device')
                if errcode == 102:
                    self.logger.info("Please reboot your device and try again")
                    raise Exception('Please reboot your device and try again')
                self.logger.error("Unknown error: %d : %s", errcode, msg[1].get("ErrorDescription", ""))
                raise Exception('Unknown error ' + str(errcode) + msg[1].get("ErrorDescription", ""))
            elif msg[0] == "DLMessageGetFreeDiskSpace":
                self.mb2_handle_free_disk_space(msg)
            elif msg[0] == "DLMessageDisconnect":
                break

    def create_status_plist(self, fullBackup=True):
        # Creating Status file for backup
        statusDict = {'UUID': str(uuid4()).upper(),
                      'BackupState': 'new',
                      'IsFullBackup': fullBackup,
                      'Version': '2.4',
                      'Date': datetime.datetime.fromtimestamp(mktime(gmtime())),
                      'SnapshotState': 'finished'
                      }
        with open(self.check_filename("Status.plist"), 'wb') as f:
            plistlib.dump(statusDict, f)

    def create_info_plist(self):
        # Get device information
        device_info = self.lockdown.all_values

        # Get a list of installed user applications
        instpxy = InstallationProxyService(self.lockdown)
        apps = instpxy.browse({"ApplicationType": "User"},
                              ["CFBundleIdentifier", "ApplicationSINF", "iTunesMetadata"])
        # Create new info.plits
        info = {"BuildVersion": device_info.get("BuildVersion") or "",
                "DeviceName": device_info.get("DeviceName") or "", "Display Name": device_info.get("DeviceName") or "",
                "GUID": "---", "Product Name": device_info.get("ProductName" or ""),
                "ProductType": device_info.get("ProductType") or "",
                "ProductVersion": device_info.get("ProductVersion") or "",
                "Serial Number": device_info.get("SerialNumber") or "", "Unique Identifier": self.udid.upper(),
                "Target Identifier": self.udid, "Target Type": "Device", "iTunes Version": "10.0.1",
                "MEID": device_info.get("MobileEquipmentIdentifier") or "",
                "Phone Number": device_info.get("PhoneNumber") or "",
                "ICCID": device_info.get("IntegratedCircuitCardIdentity") or "",
                "IMEI": device_info.get("InternationalMobileEquipmentIdentity") or "",
                "Last Backup Date": datetime.datetime.now()}

        # Starting SpringBoard service to retrieve icons position
        self.sbs = SpringBoardServicesService(self.lockdown)
        installed_apps = []
        apps_data = {}
        for app_entry in apps:
            tmp = {}
            bid = app_entry.get("CFBundleIdentifier")
            if bid:
                installed_apps.append(bid)
                pngdata = self.sbs.get_icon_pngdata(bid)
                if pngdata:
                    tmp["PlaceholderIcon"] = pngdata
                tmp["iTunesMetadata"] = app_entry.get("iTunesMetadata")
                tmp["ApplicationSINF"] = app_entry.get("ApplicationSINF")
                apps_data[bid] = tmp

        info["Applications"] = apps_data
        info["Installed Applications"] = installed_apps
        # Handling itunes files
        iTunesFiles = ["ApertureAlbumPrefs", "IC-Info.sidb", "IC-Info.sidv", "PhotosFolderAlbums",
                       "PhotosFolderName", "PhotosFolderPrefs", "VoiceMemos.plist", "iPhotoAlbumPrefs",
                       "iTunesApplicationIDs", "iTunesPrefs", "iTunesPrefs.plist"]
        iTunesFilesDict = {}
        for i in iTunesFiles:
            data = self.afc.get_file_contents("/iTunes_Control/iTunes/" + i)
            if data:
                iTunesFilesDict[i] = data

        info["iTunesFiles"] = iTunesFilesDict
        iBooksData2 = self.afc.get_file_contents("/Books/iBooksData2.plist")
        if iBooksData2:
            info["iBooks Data 2"] = iBooksData2

        info["iTunes Settings"] = self.lockdown.get_value("com.apple.iTunes")
        self.logger.info("Creating %s", os.path.join(self.udid, "Info.plist"))
        self.write_file(os.path.join(self.udid, "Info.plist"), plistlib.writePlistToString(info))

    def backup(self, full_backup=True):
        # TODO set_sync_lock
        self.logger.info("Starting %s backup...", ("Encrypted " if self.willEncrypt else ""))
        if not os.path.isdir(os.path.join(self.backupPath, self.udid)):
            os.makedirs(os.path.join(self.backupPath, self.udid))
        self.logger.info("Backup mode: %s", "Full backup" if full_backup else "Incremental backup")
        self.create_info_plist()
        options = {"ForceFullBackup": full_backup}
        self.mobilebackup2_send_request("Backup", self.udid, options)
        self.work_loop()

    def restore(self, options=None,
                password=None):

        if options is None:
            options = {"RestoreSystemFiles": True,
                       "RestoreShouldReboot": True,
                       "RestorePreserveCameraRoll": True,
                       "RemoveItemsNotRestored": False,
                       "RestoreDontCopyBackup": True,
                       "RestorePreserveSettings": True}
        self.logger.info("Starting restoration...")
        m = os.path.join(self.backupPath, self.udid, "Manifest.plist")
        try:
            with open(m, 'rb') as f:
                manifest = plistlib.load(f)
        except IOError:
            self.logger.error("not a valid backup folder")
            return -1
        if manifest.get("IsEncrypted"):
            print("Backup is encrypted, enter password : ")
            if password:
                self.password = password
            else:
                self.password = input()
            options["Password"] = self.password
        self.mobilebackup2_send_request("Restore", self.udid, self.udid, options)
        self.work_loop()

    def info(self, options=None):
        if options is None:
            options = {}
        source_udid = self.udid
        self.mobilebackup2_send_request("Info", self.udid, source_udid, options)
        self.work_loop()

    def list(self, options=None):
        if options is None:
            options = {}
        source_udid = self.udid
        self.mobilebackup2_send_request("List", self.udid, source_udid, options)
        self.work_loop()

    def changepw(self, oldpw, newpw):
        options = {"OldPassword": oldpw,
                   "NewPassword": newpw}

        self.mobilebackup2_send_request("ChangePassword", self.udid, "", options)
        self.work_loop()

    def unback(self, options=None):
        if options is None:
            options = {"Password": None}
        source_udid = self.udid
        self.mobilebackup2_send_request("Unback", self.udid, source_udid, options)
        self.work_loop()

    def enableCloudBackup(self, options=None):
        if options is None:
            options = {"CloudBackupState": False}
        self.mobilebackup2_send_request("EnableCloudBackup", self.udid, options)
        self.work_loop()


if __name__ == "__main__":
    parser = OptionParser(usage="%prog -u <udid> cmd <command options>")
    parser.add_option("-u", "--udid", default=False, action="store", dest="device_udid", metavar="DEVICE_UDID",
                      help="Device udid")
    parser.add_option("-b", "--backup", dest="backup", action="store_true", default=False,
                      help="Backup device")
    parser.add_option("-r", "--restore", dest="restore", action="store_true", default=False,
                      help="Restore device")
    parser.add_option("-i", "--info", dest="info", action="store_true", default=False,
                      help="Show backup info")
    parser.add_option("-l", "--list", dest="list", action="store_true", default=False,
                      help="Show backup info")
    parser.add_option("-p", "--path", dest="path", action="store", default=False,
                      help="path to backup/restore to")
    (options, args) = parser.parse_args()

    logging.basicConfig(level=logging.INFO)
    lockdown = LockdownClient(options.device_udid)
    mb = MobileBackup2(lockdown, options.path)
    if options.backup:
        mb.backup(full_backup=False)
    elif options.restore:
        mb.restore()
    elif options.info:
        mb.info()
    elif options.list:
        mb.list()
    else:
        parser.error("Incorrect number of arguments")



import logging

from pymobiledevice3.lockdown import LockdownClient

SRCFILES = """Baseband
CrashReporter
MobileAsset
VARFS
HFSMeta
Lockdown
MobileBackup
MobileDelete
MobileInstallation
MobileNotes
Network
UserDatabases
WiFi
WirelessAutomation
NANDDebugInfo
SystemConfiguration
Ubiquity
tmp
WirelessAutomation"""


class DeviceVersionNotSupported(Exception):
    pass


class FileRelayService(object):
    SERVICE_NAME = 'com.apple.mobile.file_relay'

    def __init__(self, lockdown: LockdownClient, ):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        self.service = self.lockdown.start_service(self.SERVICE_NAME)
        self.packet_num = 0

    def stop_session(self):
        self.logger.info("Disconecting...")
        self.service.close()

    def request_sources(self, sources=["UserDatabases"]):
        self.service.send_plist({"Sources": sources})
        while 1:
            res = self.service.recv_plist()
            if res:
                s = res.get("Status")
                if s == "Acknowledged":
                    z = ""
                    while True:
                        x = self.service.recv()
                        if not x:
                            break
                        z += x
                    return z
                else:
                    print(res.get("Error"))
                    break
        return None



import logging

from pymobiledevice3.lockdown import LockdownClient

SB_PORTRAIT = 1
SB_PORTRAIT_UPSIDE_DOWN = 2
SB_LANDSCAPE = 3
SB_LANDSCAPE_HOME_TO_LEFT = 4


class SpringBoardServicesService(object):
    SERVICE_NAME = 'com.apple.springboardservices'

    def __init__(self, lockdown: LockdownClient):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        self.service = self.lockdown.start_service(self.SERVICE_NAME)

    def get_icon_state(self, format_version="2"):
        cmd = {"command": "getIconState"}
        if format_version:
            cmd["formatVersion"] = format_version

        self.service.send_plist(cmd)
        res = self.service.recv_plist()
        return res

    def set_icon_state(self, newstate=None):
        if newstate is None:
            newstate = {}
        cmd = {"command": "setIconState",
               "iconState": newstate}

        self.service.send_plist(cmd)

    def get_icon_pngdata(self, bid):
        cmd = {"command": "getIconPNGData",
               "bundleId": bid}

        self.service.send_plist(cmd)
        res = self.service.recv_plist()
        pngdata = res.get("pngData")
        if res:
            return pngdata
        return None

    def get_interface_orientation(self):
        cmd = {"command": "getInterfaceOrientation"}
        self.service.send_plist(cmd)
        res = self.service.recv_plist()
        return res.get('interfaceOrientation')

    def get_wallpaper_pngdata(self):
        cmd = {"command": "getHomeScreenWallpaperPNGData"}
        self.service.send_plist(cmd)
        res = self.service.recv_plist()
        if res:
            return res.get("pngData")
        return None



#!/usr/bin/env python3
from functools import partial
from pprint import pprint
import logging
import tempfile
import json
import os

from pygments import highlight, lexers, formatters
from daemonize import Daemonize
from termcolor import colored
import coloredlogs
import click

from pymobiledevice3.afc import AFCShell, AFCClient
from pymobiledevice3.tcp_forwarder import TcpForwarder
from pymobiledevice3.services.diagnostics_service import DiagnosticsService
from pymobiledevice3.services.house_arrest_service import HouseArrestService
from pymobiledevice3.services.installation_proxy_service import InstallationProxyService
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.mobile_config_service import MobileConfigService
from pymobiledevice3.services.mobile_image_mounter_service import MobileImageMounterService
from pymobiledevice3.services.notification_proxy_service import NotificationProxyService
from pymobiledevice3.services.os_trace_service import OsTraceService
from pymobiledevice3.services.pcapd_service import PcapdService
from pymobiledevice3.services.screenshot_service import ScreenshotService
from pymobiledevice3.services.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.syslog_service import SyslogService

coloredlogs.install(level=logging.DEBUG)

logging.getLogger('asyncio').disabled = True
logging.getLogger('parso.cache').disabled = True
logging.getLogger('parso.cache.pickle').disabled = True
logging.getLogger('parso.python.diff').disabled = True
logging.getLogger('humanfriendly.prompts').disabled = True


def print_object(buf, colored=True):
    if colored:
        formatted_json = json.dumps(buf, sort_keys=True, indent=4)
        colorful_json = highlight(formatted_json, lexers.JsonLexer(), formatters.TerminalFormatter())
        print(colorful_json)
    else:
        pprint(buf)


class Command(click.Command):
    @staticmethod
    def udid(ctx, param, value):
        return LockdownClient(udid=value)

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        self.params[:0] = [
            click.Option(('lockdown', '--udid'), callback=self.udid),
        ]


@click.group()
def cli():
    pass


@cli.group()
def mounter():
    """ mounter options """
    pass


@mounter.command('list', cls=Command)
def mounter_list(lockdown):
    """ lookup mounter image type """
    pprint(MobileImageMounterService(lockdown=lockdown).list_images())


@mounter.command('lookup', cls=Command)
@click.argument('image_type')
def mounter_lookup(lockdown, image_type):
    """ lookup mounter image type """
    pprint(MobileImageMounterService(lockdown=lockdown).lookup_image(image_type))


@mounter.command('umount', cls=Command)
def mounter_umount(lockdown):
    """ unmount developer image. """
    image_type = 'Developer'
    mount_path = '/Developer'
    image_mounter = MobileImageMounterService(lockdown=lockdown)
    image_mounter.umount(image_type, mount_path, b'')


@mounter.command('mount', cls=Command)
@click.option('-i', '--image', type=click.Path(exists=True))
@click.option('-s', '--signature', type=click.Path(exists=True))
@click.option('-x', '--xcode', type=click.Path(exists=True, dir_okay=True, file_okay=False),
              default='/Applications/Xcode.app',
              help='Xcode application path used to figure out automatically the DeveloperDiskImage path')
@click.option('-v', '--version', help='use a different DeveloperDiskImage version from the one retrieved by lockdown'
                                      'connection')
def mounter_mount(lockdown, image, signature, xcode, version):
    """ mount developer image. """
    image_type = 'Developer'

    if image and signature:
        logging.info('using given image and signature for mount command')
    else:
        logging.info('trying to figure out the best suited DeveloperDiskImage')
        if version is None:
            version = lockdown.ios_version
        image = f'{xcode}/Contents/Developer/Platforms/iPhoneOS.platform/DeviceSupport/{version}/DeveloperDiskImage.dmg'
        signature = f'{image}.signature'

    with open(image, 'rb') as image:
        image = image.read()

    with open(signature, 'rb') as signature:
        signature = signature.read()

    image_mounter = MobileImageMounterService(lockdown=lockdown)
    image_mounter.upload_image(image_type, image, signature)
    image_mounter.mount(image_type, signature)


@cli.group()
def apps():
    """ application options """
    pass


@apps.command('list', cls=Command)
@click.option('-u', '--user', is_flag=True, help='include user apps')
@click.option('-s', '--system', is_flag=True, help='include system apps')
def apps_list(lockdown, user, system):
    """ list installed apps """
    app_types = []
    if user:
        app_types.append('User')
    if system:
        app_types.append('System')
    pprint(InstallationProxyService(lockdown=lockdown).get_apps(app_types))


@apps.command('uninstall', cls=Command)
@click.argument('bundle_id')
def apps_uninstall(lockdown, bundle_id):
    """ uninstall app by given bundle_id """
    pprint(InstallationProxyService(lockdown=lockdown).uninstall(bundle_id))


@apps.command('install', cls=Command)
@click.argument('ipa_path', type=click.Path(exists=True))
def apps_install(lockdown, ipa_path):
    """ install given .ipa """
    pprint(InstallationProxyService(lockdown=lockdown).install_from_local(ipa_path))


@apps.command('afc', cls=Command)
@click.argument('bundle_id')
def apps_afc(lockdown, bundle_id):
    """ open an AFC shell for given bundle_id, assuming its profile is installed """
    HouseArrestService(lockdown=lockdown).shell(bundle_id)


@cli.group()
def profile():
    """ profile options """
    pass


@profile.command('list', cls=Command)
def profile_list(lockdown):
    """ list installed profiles """
    pprint(MobileConfigService(lockdown=lockdown).get_profile_list())


@profile.command('install', cls=Command)
@click.argument('profile', type=click.File('rb'))
def profile_install(lockdown, profile):
    """ install given profile file """
    pprint(MobileConfigService(lockdown=lockdown).install_profile(profile.read()))


@profile.command('remove', cls=Command)
@click.argument('name')
def profile_remove(lockdown, name):
    """ remove profile by name """
    pprint(MobileConfigService(lockdown=lockdown).remove_profile(name))


@cli.group()
def lockdown():
    """ lockdown options """
    pass


@lockdown.command('forward', cls=Command)
@click.argument('src_port', type=click.IntRange(1, 0xffff))
@click.argument('dst_port', type=click.IntRange(1, 0xffff))
@click.option('-d', '--daemonize', is_flag=True)
def lockdown_forward(lockdown, src_port, dst_port, daemonize):
    """ forward tcp port """
    forwarder = TcpForwarder(lockdown, src_port, dst_port)

    if daemonize:
        with tempfile.NamedTemporaryFile('wt') as pid_file:
            daemon = Daemonize(app=f'forwarder {src_port}->{dst_port}', pid=pid_file.name, action=forwarder.start)
            daemon.start()
    else:
        forwarder.start()


@lockdown.command('recovery', cls=Command)
def lockdown_recovery(lockdown):
    """ enter recovery """
    pprint(lockdown.enter_recovery())


@lockdown.command('service', cls=Command)
@click.argument('service_name')
def lockdown_service(lockdown, service_name):
    """ send-receive raw service messages """
    lockdown.start_service(service_name).shell()


@lockdown.command('info', cls=Command)
def lockdown_info(lockdown):
    """ query all lockdown values """
    pprint(lockdown.all_values)


@cli.group()
def diagnostics():
    """ diagnostics options """
    pass


@diagnostics.command('restart', cls=Command)
def diagnostics_restart(lockdown):
    """ restart device """
    DiagnosticsService(lockdown=lockdown).restart()


@diagnostics.command('shutdown', cls=Command)
def diagnostics_shutdown(lockdown):
    """ shutdown device """
    DiagnosticsService(lockdown=lockdown).shutdown()


@diagnostics.command('sleep', cls=Command)
def diagnostics_sleep(lockdown):
    """ put device into sleep """
    DiagnosticsService(lockdown=lockdown).sleep()


@diagnostics.command('info', cls=Command)
def diagnostics_info(lockdown):
    """ get diagnostics info """
    pprint(DiagnosticsService(lockdown=lockdown).info())


@diagnostics.command('ioregistry', cls=Command)
@click.option('--plane')
@click.option('--name')
@click.option('--ioclass')
def diagnostics_ioregistry(lockdown, plane, name, ioclass):
    """ get ioregistry info """
    pprint(DiagnosticsService(lockdown=lockdown).ioregistry(plane=plane, name=name, ioclass=ioclass))


@diagnostics.command('mg', cls=Command)
@click.argument('keys', nargs=-1, default=None)
def diagnostics_mg(lockdown, keys):
    """ get MobileGestalt key values from given list. If empty, return all known. """
    pprint(DiagnosticsService(lockdown=lockdown).mobilegestalt(keys=keys))


@cli.group()
def syslog():
    """ syslog options """
    pass


@syslog.command('live-old', cls=Command)
def syslog_live_old(lockdown):
    """ view live syslog lines in raw bytes form from old relay """
    for line in SyslogService(lockdown=lockdown).watch():
        print(line)


@syslog.command('live', cls=Command)
@click.option('-o', '--out', type=click.File('wt'), help='log file')
@click.option('--nocolor', is_flag=True, help='disable colors')
@click.option('--pid', type=click.INT, default=-1, help='pid to filter. -1 for all')
@click.option('-m', '--match', help='match expression')
def syslog_live(lockdown, out, nocolor, pid, match):
    """ view live syslog lines """

    log_level_colors = {
        'Notice': 'white',
        'Error': 'red',
        'Fault': 'red',
        'Warning': 'yellow',
    }

    for syslog_entry in OsTraceService(lockdown=lockdown).syslog(pid=pid):
        pid = syslog_entry.pid
        timestamp = syslog_entry.timestamp
        level = syslog_entry.level
        filename = syslog_entry.filename
        image_name = os.path.basename(syslog_entry.image_name)
        message = syslog_entry.message
        process_name = os.path.basename(filename)

        if not nocolor:
            timestamp = colored(str(timestamp), 'green')
            process_name = colored(process_name, 'magenta')
            if len(image_name) > 0:
                image_name = colored(image_name, 'magenta')
            pid = colored(syslog_entry['pid'], 'cyan')

            if level in syslog_entry:
                level = colored(level, log_level_colors[level])

            message = colored(syslog_entry['message'], 'white')

        line = '{timestamp} {process_name}{{{image_name}}}[{pid}] <{level}>: {message}'.format(
            timestamp=timestamp, process_name=process_name, image_name=image_name, pid=pid, level=level,
            message=message,
        )

        if match and match not in line:
            continue

        print(line)

        if out:
            out.write(line)


@syslog.command('archive', cls=Command)
@click.argument('out', type=click.File('wb'))
def syslog_archive(lockdown, out):
    """
    create PAX archive.
    use `pax -r < filename` for extraction.
    """
    result, tar = OsTraceService(lockdown=lockdown).create_archive()
    out.write(tar)


@cli.command(cls=Command)
@click.argument('out', type=click.File('wb'), required=False)
def pcap(lockdown, out):
    """ sniff device traffic """
    PcapdService(lockdown=lockdown).watch(out=out)


@cli.command(cls=Command)
@click.argument('out', type=click.File('wb'))
def screenshot(lockdown, out):
    """ take a screenshot in PNG format """
    out.write(ScreenshotService(lockdown=lockdown).take_screenshot())


@cli.command(cls=Command)
@click.argument('action', type=click.Choice(['flush', 'shell']))
def crash(lockdown, action):
    """ crash utils """
    if action == 'flush':
        ack = b'ping\x00'
        assert ack == lockdown.start_service('com.apple.crashreportmover').recv_exact(len(ack))
    elif action == 'shell':
        AFCShell(lockdown=lockdown, afcname='com.apple.crashreportcopymobile').cmdloop()


@cli.group()
def afc():
    """ FileSystem utils """
    pass


@afc.command('shell', cls=Command)
def afc_shell(lockdown):
    """ open an AFC shell rooted at /var/mobile/Media """
    AFCShell(lockdown=lockdown, afcname='com.apple.afc').cmdloop()


@afc.command('pull', cls=Command)
@click.argument('remote_file', type=click.Path(exists=False))
@click.argument('local_file', type=click.File('wb'))
def afc_pull(lockdown, remote_file, local_file):
    """ open an AFC shell rooted at /var/mobile/Media """
    local_file.write(AFCClient(lockdown=lockdown).get_file_contents(remote_file))


@afc.command('push', cls=Command)
@click.argument('local_file', type=click.File('rb'))
@click.argument('remote_file', type=click.Path(exists=False))
def afc_push(lockdown, local_file, remote_file):
    """ open an AFC shell rooted at /var/mobile/Media """
    AFCClient(lockdown=lockdown).set_file_contents(remote_file, local_file.read())


@afc.command('ls', cls=Command)
@click.argument('remote_file', type=click.Path(exists=False))
def afc_ls(lockdown, remote_file):
    """ open an AFC shell rooted at /var/mobile/Media """
    pprint(AFCClient(lockdown=lockdown).read_directory(remote_file))


@afc.command('rm', cls=Command)
@click.argument('remote_file', type=click.Path(exists=False))
def afc_rm(lockdown, remote_file):
    """ open an AFC shell rooted at /var/mobile/Media """
    AFCClient(lockdown=lockdown).file_remove(remote_file)


@cli.command(cls=Command)
def ps(lockdown):
    """ show process list """
    pprint(OsTraceService(lockdown=lockdown).get_pid_list())


@cli.command(cls=Command)
@click.argument('action', type=click.Choice(['post', 'observe']))
@click.argument('names', nargs=-1)
def notification(lockdown, action, names):
    """ API for notify_post() & notify_register_dispatch(). """
    service = NotificationProxyService(lockdown=lockdown)
    for name in names:
        if action == 'post':
            service.notify_post(name)
        elif action == 'observe':
            service.notify_register_dispatch(name)

    if action == 'observe':
        for event in service.receive_notification():
            logging.info(event)


@cli.group()
def developer():
    """ developer options """
    pass


@developer.command('proclist', cls=Command)
@click.option('--nocolor', is_flag=True)
def proclist(lockdown, nocolor):
    """ show process list """
    with DvtSecureSocketProxyService(lockdown=lockdown) as dvt:
        processes = dvt.proclist()
        for process in processes:
            if 'startDate' in process:
                process['startDate'] = str(process['startDate'])

        print_object(processes, colored=not nocolor)


@developer.command('applist', cls=Command)
@click.option('--nocolor', is_flag=True)
def applist(lockdown, nocolor):
    """ show application list """
    with DvtSecureSocketProxyService(lockdown=lockdown) as dvt:
        apps = dvt.applist()
        print_object(apps, colored=not nocolor)


@developer.command('kill', cls=Command)
@click.argument('pid', type=click.INT)
def kill(lockdown, pid):
    """ Kill a process by its pid. """
    with DvtSecureSocketProxyService(lockdown=lockdown) as dvt:
        dvt.kill(pid)


@developer.command('launch', cls=Command)
@click.argument('bundle_id', type=click.STRING)
def launch(lockdown, bundle_id):
    """ Kill a process by its pid. """
    with DvtSecureSocketProxyService(lockdown=lockdown) as dvt:
        pid = dvt.launch(bundle_id)
        print(f'Process launched with pid {pid}')


@developer.command('shell', cls=Command)
def shell(lockdown):
    """ Launch developer shell. """
    with DvtSecureSocketProxyService(lockdown=lockdown) as dvt:
        dvt.shell()


@developer.command('ls', cls=Command)
@click.argument('path', type=click.Path(exists=False))
def ls(lockdown, path):
    """ List directory. """
    with DvtSecureSocketProxyService(lockdown=lockdown) as dvt:
        pprint(dvt.ls(path))


@developer.command('device-information', cls=Command)
@click.option('--nocolor', is_flag=True)
def device_information(lockdown, nocolor):
    """ Print system information. """
    with DvtSecureSocketProxyService(lockdown=lockdown) as dvt:
        print_object({
            'system': dvt.system_information(),
            'hardware': dvt.hardware_information(),
            'network': dvt.network_information(),
        }, colored=not nocolor)


if __name__ == '__main__':
    cli()



#!/usr/bin/env python3

"""
How to create a CA certificate with Python.

WARNING: This sample only demonstrates how to use the objects and methods,
         not how to create a safe and correct certificate.

Copyright (c) 2004 Open Source Applications Foundation.
Authors: Heikki Toivonen
         Mathieu RENARD
"""
import base64

from M2Crypto import RSA, X509, EVP, m2, BIO
from M2Crypto.RSA import load_pub_key_bio
from pyasn1.type import univ
from pyasn1.codec.der import encoder as der_encoder
from pyasn1.codec.der import decoder as der_decoder


def convert_pkcs1_to_pkcs8_pubkey(bitsdata):
    pubkey_pkcs1_b64 = b''.join(bitsdata.split(b'\n')[1:-2])
    pubkey_pkcs1, restOfInput = der_decoder.decode(base64.b64decode(pubkey_pkcs1_b64))
    bitstring = univ.Sequence()
    bitstring.setComponentByPosition(0, univ.Integer(pubkey_pkcs1[0]))
    bitstring.setComponentByPosition(1, univ.Integer(pubkey_pkcs1[1]))
    bitstring = der_encoder.encode(bitstring)
    try:
        bitstring = ''.join([('00000000' + bin(ord(x))[2:])[-8:] for x in list(bitstring)])
    except:
        bitstring = ''.join([('00000000' + bin(x)[2:])[-8:] for x in list(bitstring)])
    bitstring = univ.BitString("'%s'B" % bitstring)
    pubkeyid = univ.Sequence()
    pubkeyid.setComponentByPosition(0, univ.ObjectIdentifier('1.2.840.113549.1.1.1'))  # == OID for rsaEncryption
    pubkeyid.setComponentByPosition(1, univ.Null(''))
    pubkey_seq = univ.Sequence()
    pubkey_seq.setComponentByPosition(0, pubkeyid)
    pubkey_seq.setComponentByPosition(1, bitstring)
    base64.MAXBINSIZE = (64 // 4) * 3
    res = b"-----BEGIN PUBLIC KEY-----\n"
    res += base64.b64encode(der_encoder.encode(pubkey_seq))
    res += b"-----END PUBLIC KEY-----\n"
    return res


def generate_rsa_key():
    return RSA.gen_key(2048, m2.RSA_F4)


def make_pkey(key):
    pkey = EVP.PKey()
    pkey.assign_rsa(key)
    return pkey


def make_request(pkey, cn):
    req = X509.Request()
    req.set_version(2)
    req.set_pubkey(pkey)
    name = X509.X509_Name()
    name.CN = cn
    req.set_subject_name(name)
    ext1 = X509.new_extension('subjectAltName', 'DNS:foobar.example.com')
    ext2 = X509.new_extension('nsComment', 'Hello there')
    extstack = X509.X509_Extension_Stack()
    extstack.push(ext1)
    extstack.push(ext2)

    assert (extstack[1].get_name() == 'nsComment')

    req.add_extensions(extstack)
    return req


def make_cert(req, caPkey):
    pkey = req.get_pubkey()
    # if not req.verify(woop.pkey):
    if not req.verify(pkey):
        # XXX What error object should I use?
        raise ValueError('Error verifying request')
    sub = req.get_subject()
    # If this were a real certificate request, you would display
    # all the relevant data from the request and ask a human operator
    # if you were sure. Now we just create the certificate blindly based
    # on the request.
    cert = X509.X509()
    # We know we are making CA cert now...
    # Serial defaults to 0.
    cert.set_serial_number(1)
    cert.set_version(2)
    cert.set_subject(sub)
    issuer = X509.X509_Name()
    issuer.CN = 'The Issuer Monkey'
    issuer.O = 'The Organization Otherwise Known as My CA, Inc.'
    cert.set_issuer(issuer)
    cert.set_pubkey(pkey)
    notBefore = m2.x509_get_not_before(cert.x509)
    notAfter = m2.x509_get_not_after(cert.x509)
    m2.x509_gmtime_adj(notBefore, 0)
    days = 30
    m2.x509_gmtime_adj(notAfter, 60 * 60 * 24 * days)
    cert.add_ext(
        X509.new_extension('subjectAltName', 'DNS:foobar.example.com'))
    ext = X509.new_extension('nsComment', 'M2Crypto generated certificate')
    ext.set_critical(0)  # Defaults to non-critical, but we can also set it
    cert.add_ext(ext)
    cert.sign(caPkey, 'sha1')

    assert (cert.get_ext('subjectAltName').get_name() == 'subjectAltName')
    assert (cert.get_ext_at(0).get_name() == 'subjectAltName')
    assert (cert.get_ext_at(0).get_value() == 'DNS:foobar.example.com')

    return cert


def ca():
    key = generate_rsa_key()
    pkey = make_pkey(key)
    req = make_request(pkey)
    cert = make_cert(req, pkey)
    return (cert, pkey)


def ca_do_everything(DevicePublicKey):
    rsa = generate_rsa_key()
    privateKey = make_pkey(rsa)
    req = make_request(privateKey, "The Issuer Monkey")
    cert = make_cert(req, privateKey)
    rsa2 = load_pub_key_bio(BIO.MemoryBuffer(
        convert_pkcs1_to_pkcs8_pubkey(DevicePublicKey)))
    pkey2 = EVP.PKey()
    pkey2.assign_rsa(rsa2)
    req = make_request(pkey2, "Device")
    cert2 = make_cert(req, privateKey)
    return cert.as_pem(), privateKey.as_pem(None), cert2.as_pem()


if __name__ == '__main__':
    rsa = generate_rsa_key()
    pkey = make_pkey(rsa)
    print(pkey.as_pem(None))
    req = make_request(pkey, "The Issuer Monkey")
    cert = make_cert(req, pkey)
    print(cert.as_text())
    cert.save_pem('my_ca_cert.pem')
    rsa.save_key('my_key.pem', None)
